/*
** 2000-05-29
**
** The author disclaims copyright to this source code.  In place of
** a legal notice, here is a blessing:
**
**    May you do good and not evil.
**    May you find forgiveness for yourself and forgive others.
**    May you share freely, never taking more than you give.
**
*************************************************************************
** Driver template for the LEMON parser generator.
**
** The "lemon" program processes an LALR(1) input grammar file, then uses
** this template to construct a parser.  The "lemon" program inserts text
** at each "%%" line.  Also, any "P-a-r-s-e" identifer prefix (without the
** interstitial "-" characters) contained in this template is changed into
** the value of the %name directive from the grammar.  Otherwise, the content
** of this template is copied straight through into the generate parser
** source file.
**
** The following is the concatenation of all %include directives from the
** input grammar file:
*/
#include <stdio.h>
/************ Begin %include sections from the grammar ************************/
#line 1 "./mate_grammar.lemon"


/* mate_grammar.lemon
* MATE's configuration language grammar
*
* Copyright 2005, Luis E. Garcia Ontanon <luis@ontanon.org>
*
* Wireshark - Network traffic analyzer
* By Gerald Combs <gerald@wireshark.org>
* Copyright 1998 Gerald Combs
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include "mate.h"
#include "mate_grammar.h"
#include <wsutil/file_util.h>

#define DUMMY void*

typedef struct _extraction {
	gchar* as;
	header_field_info* hfi;
	struct _extraction* next;
	struct _extraction* last;
} extraction_t;

typedef struct _pdu_criteria_t {
	AVPL* criterium_avpl;
	avpl_match_mode criterium_match_mode;
	accept_mode_t criterium_accept_mode;
} pdu_criteria_t;

typedef struct _gop_options {
	gop_tree_mode_t pdu_tree_mode;
	gboolean drop_unassigned;
	gboolean show_times;
	float expiration;
	float idle_timeout;
	float lifetime;
	AVPL* start;
	AVPL* stop;
	AVPL* extras;
} gop_options_t;

typedef struct _gog_statements {
	float expiration;
	gop_tree_mode_t gop_tree_mode;
	GPtrArray* transform_list;
	AVPL* extras;
	LoAL* current_gogkeys;
} gog_statement_t;

typedef struct _transf_match_t {
	avpl_match_mode match_mode;
	AVPL* avpl;
} transf_match_t;

typedef struct _transf_action_t {
	avpl_replace_mode replace_mode;
	AVPL* avpl;
} transf_action_t;

static void configuration_error(mate_config* mc, const gchar* fmt, ...) {
	static gchar error_buffer[256];
	const gchar* incl;
	gint i;
	mate_config_frame* current_frame;
	va_list list;

	va_start( list, fmt );
	g_vsnprintf(error_buffer,sizeof(error_buffer),fmt,list);
	va_end( list );

	i = (gint) mc->config_stack->len;

	while (i--) {

		if (i>0) {
			incl = "\n   included from: ";
		} else {
			incl = " ";
		}

		current_frame = (mate_config_frame *)g_ptr_array_index(mc->config_stack,(guint)i);

		g_string_append_printf(mc->config_error,"%s%s at line %u",incl, current_frame->filename, current_frame->linenum);
	}

	g_string_append_printf(mc->config_error,": %s\n",error_buffer);

	THROW(MateConfigError);

}

static AVPL_Transf* new_transform_elem(AVPL* match, AVPL* replace, avpl_match_mode match_mode, avpl_replace_mode replace_mode) {
	 AVPL_Transf* t = (AVPL_Transf *)g_malloc(sizeof(AVPL_Transf));

	 t->name = NULL;
	 t->match = match;
	 t->replace = replace;
	 t->match_mode = match_mode;
	 t->replace_mode = replace_mode;

	 t->map = NULL;
	 t->next = NULL;

	 return t;
}

static gchar* recolonize(mate_config* mc, gchar* s) {
	GString* str = g_string_new("");
	gchar** vec;
	gchar* r;
	guint i,v;
	gchar c;

	vec = g_strsplit(s,":",0);

	for (i = 0; vec[i]; i++) {
		g_ascii_strdown(vec[i], -1);

		v = 0;
		switch ( strlen(vec[i]) ) {
		 case 2:
			c = vec[i][1];
			vec[i][1] = vec[i][0];
			vec[i][0] = c;
			if (vec[i][0] >= '0' && vec[i][0] <= '9') {
				v += (vec[i][1] - '0' )*16;
			} else {
				v += (vec[i][1] - 'a' + 10)*16;
			}
		 case 1:
			if (vec[i][0] >= '0' && vec[i][0] <= '9') {
				v += (vec[i][0] - '0' );
			} else {
				v += (vec[i][0] - 'a' + 10);
			}
		 case 0:
			break;
		  default:
			configuration_error(mc,"bad token %s",s);
		}

		g_string_append_printf(str,":%.2X",v);
	}

	g_strfreev(vec);

	g_string_erase(str,0,1);

	r = str->str;

	g_string_free(str,FALSE);

	return r;
}

#line 200 "mate_grammar.c"
/**************** End of %include directives **********************************/
/* These constants specify the various numeric values for terminal symbols
** in a format understandable to "makeheaders".  This section is blank unless
** "lemon" is run with the "-m" command-line option.
***************** Begin makeheaders token definitions *************************/
/**************** End makeheaders token definitions ***************************/

/* The next sections is a series of control #defines.
** various aspects of the generated parser.
**    YYCODETYPE         is the data type used to store the integer codes
**                       that represent terminal and non-terminal symbols.
**                       "unsigned char" is used if there are fewer than
**                       256 symbols.  Larger types otherwise.
**    YYNOCODE           is a number of type YYCODETYPE that is not used for
**                       any terminal or nonterminal symbol.
**    YYFALLBACK         If defined, this indicates that one or more tokens
**                       (also known as: "terminal symbols") have fall-back
**                       values which should be used if the original symbol
**                       would not parse.  This permits keywords to sometimes
**                       be used as identifiers, for example.
**    YYACTIONTYPE       is the data type used for "action codes" - numbers
**                       that indicate what to do in response to the next
**                       token.
**    MateParserTOKENTYPE     is the data type used for minor type for terminal
**                       symbols.  Background: A "minor type" is a semantic
**                       value associated with a terminal or non-terminal
**                       symbols.  For example, for an "ID" terminal symbol,
**                       the minor type might be the name of the identifier.
**                       Each non-terminal can have a different minor type.
**                       Terminal symbols all have the same minor type, though.
**                       This macros defines the minor type for terminal
**                       symbols.
**    YYMINORTYPE        is the data type used for all minor types.
**                       This is typically a union of many types, one of
**                       which is MateParserTOKENTYPE.  The entry in the union
**                       for terminal symbols is called "yy0".
**    YYSTACKDEPTH       is the maximum depth of the parser's stack.  If
**                       zero the stack is dynamically sized using realloc()
**    MateParserARG_SDECL     A static variable declaration for the %extra_argument
**    MateParserARG_PDECL     A parameter declaration for the %extra_argument
**    MateParserARG_STORE     Code to store %extra_argument into yypParser
**    MateParserARG_FETCH     Code to extract %extra_argument from yypParser
**    YYERRORSYMBOL      is the code number of the error symbol.  If not
**                       defined, then do no error processing.
**    YYNSTATE           the combined number of states.
**    YYNRULE            the number of rules in the grammar
**    YY_MAX_SHIFT       Maximum value for shift actions
**    YY_MIN_SHIFTREDUCE Minimum value for shift-reduce actions
**    YY_MAX_SHIFTREDUCE Maximum value for shift-reduce actions
**    YY_MIN_REDUCE      Maximum value for reduce actions
**    YY_ERROR_ACTION    The yy_action[] code for syntax error
**    YY_ACCEPT_ACTION   The yy_action[] code for accept
**    YY_NO_ACTION       The yy_action[] code for no-op
*/
#ifndef INTERFACE
# define INTERFACE 1
#endif
/************* Begin control #defines *****************************************/
#define YYCODETYPE unsigned char
#define YYNOCODE 139
#define YYACTIONTYPE unsigned short int
#define MateParserTOKENTYPE  gchar* 
typedef union {
  int yyinit;
  MateParserTOKENTYPE yy0;
  transf_match_t* yy16;
  pdu_criteria_t* yy41;
  AVPL_Transf* yy71;
  transf_action_t* yy77;
  float yy83;
  gchar* yy88;
  gop_tree_mode_t yy93;
  AVP* yy110;
  AVPL* yy170;
  gog_statement_t* yy171;
  avpl_match_mode yy172;
  accept_mode_t yy188;
  header_field_info* yy202;
  gop_options_t* yy203;
  avpl_replace_mode yy215;
  gboolean yy216;
  GPtrArray* yy231;
  LoAL* yy243;
  extraction_t* yy259;
} YYMINORTYPE;
#ifndef YYSTACKDEPTH
#define YYSTACKDEPTH 100
#endif
#define MateParserARG_SDECL  mate_config* mc ;
#define MateParserARG_PDECL , mate_config* mc 
#define MateParserARG_FETCH  mate_config* mc  = yypParser->mc 
#define MateParserARG_STORE yypParser->mc  = mc 
#define YYNSTATE             183
#define YYNRULE              149
#define YY_MAX_SHIFT         182
#define YY_MIN_SHIFTREDUCE   287
#define YY_MAX_SHIFTREDUCE   435
#define YY_MIN_REDUCE        436
#define YY_MAX_REDUCE        584
#define YY_ERROR_ACTION      585
#define YY_ACCEPT_ACTION     586
#define YY_NO_ACTION         587
/************* End control #defines *******************************************/

/* The yyzerominor constant is used to initialize instances of
** YYMINORTYPE objects to zero. */
static const YYMINORTYPE yyzerominor = { 0 };

/* Define the yytestcase() macro to be a no-op if is not already defined
** otherwise.
**
** Applications can choose to define yytestcase() in the %include section
** to a macro that can assist in verifying code coverage.  For production
** code the yytestcase() macro should be turned off.  But it is useful
** for testing.
*/
#ifndef yytestcase
# define yytestcase(X)
#endif


/* Next are the tables used to determine what action to take based on the
** current state and lookahead token.  These tables are used to implement
** functions that take a state number and lookahead value and return an
** action integer.
**
** Suppose the action integer is N.  Then the action is determined as
** follows
**
**   0 <= N <= YY_MAX_SHIFT             Shift N.  That is, push the lookahead
**                                      token onto the stack and goto state N.
**
**   N between YY_MIN_SHIFTREDUCE       Shift to an arbitrary state then
**     and YY_MAX_SHIFTREDUCE           reduce by rule N-YY_MIN_SHIFTREDUCE.
**
**   N between YY_MIN_REDUCE            Reduce by rule N-YY_MIN_REDUCE
**     and YY_MAX_REDUCE

**   N == YY_ERROR_ACTION               A syntax error has occurred.
**
**   N == YY_ACCEPT_ACTION              The parser accepts its input.
**
**   N == YY_NO_ACTION                  No such action.  Denotes unused
**                                      slots in the yy_action[] table.
**
** The action table is constructed as a single large table named yy_action[].
** Given state S and lookahead X, the action is computed as
**
**      yy_action[ yy_shift_ofst[S] + X ]
**
** If the index value yy_shift_ofst[S]+X is out of range or if the value
** yy_lookahead[yy_shift_ofst[S]+X] is not equal to X or if yy_shift_ofst[S]
** is equal to YY_SHIFT_USE_DFLT, it means that the action is not in the table
** and that yy_default[S] should be used instead.
**
** The formula above is for computing the action when the lookahead is
** a terminal symbol.  If the lookahead is a non-terminal (as occurs after
** a reduce action) then the yy_reduce_ofst[] array is used in place of
** the yy_shift_ofst[] array and YY_REDUCE_USE_DFLT is used in place of
** YY_SHIFT_USE_DFLT.
**
** The following are the tables generated in this section:
**
**  yy_action[]        A single table containing all actions.
**  yy_lookahead[]     A table containing the lookahead for each entry in
**                     yy_action.  Used to detect hash collisions.
**  yy_shift_ofst[]    For each state, the offset into yy_action for
**                     shifting terminals.
**  yy_reduce_ofst[]   For each state, the offset into yy_action for
**                     shifting non-terminals after a reduce.
**  yy_default[]       Default action for each state.
**
*********** Begin parsing tables **********************************************/
#define YY_ACTTAB_COUNT (306)
static const YYACTIONTYPE yy_action[] = {
 /*     0 */   436,  182,  293,  181,  430,  431,   20,  433,    8,  343,
 /*    10 */     4,  138,  126,  111,  165,  393,  394,  395,  396,  353,
 /*    20 */   354,  409,  586,    1,  141,  407,  408,  350,  351,  352,
 /*    30 */   429,  101,  139,   47,   75,   43,   50,   11,  102,  103,
 /*    40 */   424,  341,  364,  401,  410,   71,  504,  432,    5,  411,
 /*    50 */   129,  288,  290,  291,  292,  294,  295,  434,  435,   70,
 /*    60 */   358,   20,  342,    4,   98,  374,  419,  365,    6,  397,
 /*    70 */   398,   99,  100,  374,  362,  363,  416,  427,   74,    3,
 /*    80 */   421,   63,   43,  400,  167,  166,  375,  422,  420,  129,
 /*    90 */   119,  428,  425,  161,  107,   66,   73,  106,   63,  157,
 /*   100 */   115,   19,    7,  116,  117,  413,   27,  105,   18,  118,
 /*   110 */   120,  109,  131,  132,   22,  133,   42,  142,  143,  144,
 /*   120 */   148,  149,  150,   11,   16,  112,  151,  156,  152,   64,
 /*   130 */    26,   14,  158,   15,   23,   25,   24,   50,   49,  399,
 /*   140 */    48,   31,   44,  124,   30,   29,   28,  102,   45,   52,
 /*   150 */   122,   51,   10,   57,   46,   55,   53,   54,   58,    2,
 /*   160 */    59,   34,   69,   33,  113,  114,   56,  135,   67,   68,
 /*   170 */   418,  140,   32,   39,  123,   17,   38,   35,   37,   21,
 /*   180 */    36,  162,  373,   40,  127,   41,  155,  130,  147,   12,
 /*   190 */   172,  178,  175,  136,  169,  104,  412,  345,  405,  348,
 /*   200 */   423,  403,  402,  108,   65,    9,  379,   60,  381,  414,
 /*   210 */   110,  387,  389,  391,  377,  383,  385,  376,   61,   13,
 /*   220 */   366,  121,   62,   72,  360,  125,  344,   78,  146,  128,
 /*   230 */   367,   77,  369,  371,  356,  340,  168,  137,   81,  154,
 /*   240 */    82,   80,   76,   83,   93,   91,   86,   84,  160,  164,
 /*   250 */    90,   95,  180,   87,   94,   96,   88,  134,  312,  314,
 /*   260 */   316,  310,  171,  320,  170,  145,   79,  305,  322,  324,
 /*   270 */   438,  326,  330,  173,  318,  438,  307,  174,  153,   85,
 /*   280 */   297,  334,  336,  438,  338,  332,  179,  176,  328,  438,
 /*   290 */   159,   97,  438,   89,  296,   92,  309,  177,  299,  438,
 /*   300 */   438,  298,  163,  438,  301,  303,
};
static const YYCODETYPE yy_lookahead[] = {
 /*     0 */     0,    1,   63,    3,    7,    8,   25,   10,   65,   66,
 /*    10 */    67,   11,   12,   13,   14,   44,   45,   46,   47,   29,
 /*    20 */    30,   47,  110,  111,   24,   51,   52,   26,   27,   28,
 /*    30 */   108,  109,   68,   86,   70,   53,   54,   24,    8,  106,
 /*    40 */   107,    5,   81,   10,   97,    4,   56,   50,   96,   97,
 /*    50 */    37,  112,  113,  114,  115,  116,  117,   60,   61,   98,
 /*    60 */     2,   25,   66,   67,  102,  103,   99,   81,   82,   48,
 /*    70 */    49,  104,  102,  103,   35,   36,    2,    5,   55,    4,
 /*    80 */    57,   40,   53,   50,    7,    8,    8,   57,    8,   37,
 /*    90 */    74,  108,  108,   74,  100,   69,   69,   72,   40,  100,
 /*   100 */    72,   77,   56,   72,   72,    8,   22,  105,   23,   75,
 /*   110 */    75,   89,   75,   75,   18,   75,   71,   75,   75,   75,
 /*   120 */    72,   72,   72,   24,   21,  105,   75,   72,   75,   55,
 /*   130 */    16,   59,   75,   58,   18,   20,   19,   54,   43,    8,
 /*   140 */    42,   39,   95,   73,   17,   16,   34,    8,   94,   87,
 /*   150 */    89,   88,   33,   86,   98,   90,   92,   91,   85,    4,
 /*   160 */    84,   17,   76,   16,  105,  105,   98,   80,   79,   78,
 /*   170 */    99,   64,   15,   22,  105,   21,   16,   18,   20,   23,
 /*   180 */    19,   13,  103,   18,  103,   22,   12,  105,   11,  101,
 /*   190 */    11,   13,   12,  103,    9,    6,    2,  105,    2,  105,
 /*   200 */   107,    2,    2,    5,   38,    4,    2,    4,    2,    2,
 /*   210 */     8,    2,    2,    2,    2,    2,    2,    2,   25,   32,
 /*   220 */     2,    5,   41,   31,    2,    8,    2,  126,  128,    8,
 /*   230 */     2,  127,    2,    2,    2,    2,   10,    8,  132,  134,
 /*   240 */   131,  133,    8,  130,  121,  123,  136,  129,  137,  125,
 /*   250 */   124,  119,  122,  133,  120,  118,  135,    5,    2,    2,
 /*   260 */     2,    2,    9,    2,   10,    5,    4,    2,    2,    2,
 /*   270 */   138,    2,    2,   10,    2,  138,    2,    9,    5,    4,
 /*   280 */     2,    2,    2,  138,    2,    2,    5,   10,    2,  138,
 /*   290 */     5,    4,  138,    4,    2,    4,    2,    9,    2,  138,
 /*   300 */   138,    2,    5,  138,    2,    2,
};
#define YY_SHIFT_USE_DFLT (-30)
#define YY_SHIFT_COUNT (182)
#define YY_SHIFT_MIN   (-29)
#define YY_SHIFT_MAX   (303)
static const short yy_shift_ofst[] = {
 /*     0 */   -30,    0,  -19,   -3,  -10,  -18,   13,   30,   36,   29,
 /*    10 */    78,   80,   52,   78,   -3,   -3,  -29,  -29,  -26,    1,
 /*    20 */     1,  -26,   33,   33,   33,   33,   21,   21,   39,   21,
 /*    30 */    21,   21,   21,   21,   21,   33,   33,   33,   21,   21,
 /*    40 */    33,   21,   46,   97,   84,   85,   96,   99,   46,   46,
 /*    50 */    46,   84,  103,  114,  115,  117,  116,   99,   83,   95,
 /*    60 */    98,   46,  131,   78,   80,   78,   46,  102,  127,  129,
 /*    70 */   112,  119,   78,   46,  139,   46,  155,  144,  147,  157,
 /*    80 */   151,  154,  160,  158,  161,  159,  163,  156,  154,  165,
 /*    90 */   168,  174,  177,  178,  180,  179,  185,  189,   58,   74,
 /*   100 */    41,   72,   75,   23,   77,  194,  196,  199,  200,  198,
 /*   110 */   201,  202,  204,  206,  207,  209,  210,  211,  212,  213,
 /*   120 */   214,  215,  216,  203,  193,  181,  217,  218,  166,  221,
 /*   130 */   222,  228,  230,  231,  232,  252,  187,  192,  229,  224,
 /*   140 */   233,  234,  256,  257,  258,  259,  260,  262,  261,  266,
 /*   150 */   267,  269,  270,  272,  273,  275,  279,  280,  282,  283,
 /*   160 */   285,  286,  289,  294,  297,  291,  296,  299,  302,  226,
 /*   170 */   303,  254,  253,  265,  263,  268,  274,  277,  288,  278,
 /*   180 */   281,  287,  292,
};
#define YY_REDUCE_USE_DFLT (-89)
#define YY_REDUCE_COUNT (97)
#define YY_REDUCE_MIN   (-88)
#define YY_REDUCE_MAX   (137)
static const short yy_reduce_ofst[] = {
 /*     0 */   -88,  -61,  -57,  -78,  -36,  -53,  -39,  -67,   -4,  -48,
 /*    10 */   -38,  -33,  -14,  -30,  -17,  -16,   16,   19,   -6,   26,
 /*    20 */    27,   -1,   25,   28,   31,   32,   34,   35,   24,   37,
 /*    30 */    38,   40,   42,   43,   44,   48,   49,   50,   51,   53,
 /*    40 */    55,   57,    2,   45,   22,   47,   54,   56,   20,   59,
 /*    50 */    60,   61,   63,   62,   64,   66,   65,   68,   67,   73,
 /*    60 */    76,   69,   70,   79,   71,   81,   82,   87,   89,   91,
 /*    70 */    86,   88,   90,   92,   93,   94,  107,  100,  104,  101,
 /*    80 */   105,  108,  106,  109,  113,  118,  111,  110,  120,  121,
 /*    90 */   124,  126,  122,  130,  123,  134,  132,  137,
};
static const YYACTIONTYPE yy_default[] = {
 /*     0 */   438,  585,  495,  585,  496,  564,  566,  585,  495,  585,
 /*    10 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  498,
 /*    20 */   498,  585,  585,  585,  585,  585,  585,  585,  510,  585,
 /*    30 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*    40 */   585,  585,  585,  585,  535,  553,  555,  566,  585,  585,
 /*    50 */   585,  535,  533,  527,  541,  539,  537,  566,  564,  531,
 /*    60 */   529,  585,  585,  585,  585,  585,  585,  521,  519,  517,
 /*    70 */   508,  506,  585,  585,  585,  585,  585,  466,  464,  462,
 /*    80 */   480,  478,  476,  474,  472,  470,  488,  486,  478,  484,
 /*    90 */   482,  468,  460,  457,  455,  453,  451,  449,  585,  585,
 /*   100 */   585,  585,  575,  585,  585,  585,  585,  585,  585,  585,
 /*   110 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*   120 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*   130 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*   140 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*   150 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*   160 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*   170 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*   180 */   585,  585,  585,
};
/********** End of lemon-generated parsing tables *****************************/

/* The next table maps tokens (terminal symbols) into fallback tokens.
** If a construct like the following:
**
**      %fallback ID X Y Z.
**
** appears in the grammar, then ID becomes a fallback token for X, Y,
** and Z.  Whenever one of the tokens X, Y, or Z is input to the parser
** but it does not parse, the type of the token is changed to ID and
** the parse is retried before an error is thrown.
**
** This feature can be used, for example, to cause some keywords in a language
** to revert to identifiers if they keyword does not apply in the context where
** it appears.
*/
#ifdef YYFALLBACK
static const YYCODETYPE yyFallback[] = {
};
#endif /* YYFALLBACK */

/* The following structure represents a single element of the
** parser's stack.  Information stored includes:
**
**   +  The state number for the parser at this level of the stack.
**
**   +  The value of the token stored at this level of the stack.
**      (In other words, the "major" token.)
**
**   +  The semantic value stored at this level of the stack.  This is
**      the information used by the action routines in the grammar.
**      It is sometimes called the "minor" token.
**
** After the "shift" half of a SHIFTREDUCE action, the stateno field
** actually contains the reduce action for the second half of the
** SHIFTREDUCE.
*/
struct yyStackEntry {
  YYACTIONTYPE stateno;  /* The state-number, or reduce action in SHIFTREDUCE */
  YYCODETYPE major;      /* The major token value.  This is the code
                         ** number for the token at this stack level */
  YYMINORTYPE minor;     /* The user-supplied minor token value.  This
                         ** is the value of the token  */
};
typedef struct yyStackEntry yyStackEntry;

/* The state of the parser is completely contained in an instance of
** the following structure */
struct yyParser {
  int yyidx;                    /* Index of top element in stack */
#ifdef YYTRACKMAXSTACKDEPTH
  int yyidxMax;                 /* Maximum value of yyidx */
#endif
  int yyerrcnt;                 /* Shifts left before out of the error */
  MateParserARG_SDECL                /* A place to hold %extra_argument */
#if YYSTACKDEPTH<=0
  int yystksz;                  /* Current side of the stack */
  yyStackEntry *yystack;        /* The parser's stack */
#else
  yyStackEntry yystack[YYSTACKDEPTH];  /* The parser's stack */
#endif
};
typedef struct yyParser yyParser;

#ifndef NDEBUG
#include <stdio.h>
static FILE *yyTraceFILE = 0;
static char *yyTracePrompt = 0;
#endif /* NDEBUG */

#ifndef NDEBUG
/*
** Turn parser tracing on by giving a stream to which to write the trace
** and a prompt to preface each trace message.  Tracing is turned off
** by making either argument NULL
**
** Inputs:
** <ul>
** <li> A FILE* to which trace output should be written.
**      If NULL, then tracing is turned off.
** <li> A prefix string written at the beginning of every
**      line of trace output.  If NULL, then tracing is
**      turned off.
** </ul>
**
** Outputs:
** None.
*/
void MateParserTrace(FILE *TraceFILE, char *zTracePrompt){
  yyTraceFILE = TraceFILE;
  yyTracePrompt = zTracePrompt;
  if( yyTraceFILE==0 ) yyTracePrompt = 0;
  else if( yyTracePrompt==0 ) yyTraceFILE = 0;
}
#endif /* NDEBUG */

#ifndef NDEBUG
/* For tracing shifts, the names of all terminals and nonterminals
** are required.  The following table supplies these names */
static const char *const yyTokenName[] = {
  "$",             "DONE_KW",       "SEMICOLON",     "DEBUG_KW",    
  "OPEN_BRACE",    "CLOSE_BRACE",   "FILENAME_KW",   "QUOTED",      
  "NAME",          "LEVEL_KW",      "INTEGER",       "PDU_KW",      
  "GOP_KW",        "GOG_KW",        "DEFAULT_KW",    "LAST_EXTRACTED_KW",
  "DROP_UNASSIGNED_KW",  "DISCARD_PDU_DATA_KW",  "EXPIRATION_KW",  "IDLE_TIMEOUT_KW",
  "LIFETIME_KW",   "SHOW_TREE_KW",  "SHOW_TIMES_KW",  "GOP_TREE_KW", 
  "TRANSFORM_KW",  "MATCH_KW",      "STRICT_KW",     "EVERY_KW",    
  "LOOSE_KW",      "REPLACE_KW",    "INSERT_KW",     "PROTO_KW",    
  "TRANSPORT_KW",  "PAYLOAD_KW",    "CRITERIA_KW",   "ACCEPT_KW",   
  "REJECT_KW",     "EXTRACT_KW",    "FROM_KW",       "LAST_PDU_KW", 
  "SLASH",         "ON_KW",         "START_KW",      "STOP_KW",     
  "NO_TREE_KW",    "PDU_TREE_KW",   "FRAME_TREE_KW",  "BASIC_TREE_KW",
  "TRUE_KW",       "FALSE_KW",      "FLOATING",      "NULL_TREE_KW",
  "FULL_TREE_KW",  "MEMBER_KW",     "EXTRA_KW",      "COMMA",       
  "OPEN_PARENS",   "CLOSE_PARENS",  "AVP_OPERATOR",  "PIPE",        
  "DOTED_IP",      "COLONIZED",     "error",         "transform_decl",
  "transform_body",  "transform_statements",  "transform_statement",  "transform_match",
  "transform_action",  "match_mode",    "action_mode",   "gop_name",    
  "time_value",    "pdu_name",      "gop_tree_mode",  "true_false",  
  "criteria_statement",  "accept_mode",   "pdu_drop_unassigned_statement",  "discard_pdu_data_statement",
  "last_extracted_statement",  "extraction_statement",  "extraction_statements",  "gop_options", 
  "gop_start_statement",  "gop_stop_statement",  "extra_statement",  "gop_drop_unassigned_statement",
  "show_goptree_statement",  "show_times_statement",  "gop_expiration_statement",  "idle_timeout_statement",
  "lifetime_statement",  "gog_statements",  "gog_expiration_statement",  "gog_goptree_statement",
  "gog_key_statements",  "gog_key_statement",  "transform_list_statement",  "transform",   
  "gop_tree_type",  "payload_statement",  "proto_stack",   "field",       
  "transform_list",  "avpl",          "avps",          "avp",         
  "value",         "avp_oneoff",    "mate_config",   "decls",       
  "decl",          "pdu_decl",      "gop_decl",      "gog_decl",    
  "defaults_decl",  "debug_decl",    "dbgfile_default",  "dbglevel_default",
  "pdu_dbglevel_default",  "gop_dbglevel_default",  "gog_dbglevel_default",  "pdu_defaults",
  "gop_defaults",  "gog_defaults",  "pdu_last_extracted_default",  "pdu_drop_unassigned_default",
  "pdu_discard_default",  "gop_expiration_default",  "gop_idle_timeout_default",  "gop_lifetime_default",
  "gop_drop_unassigned_default",  "gop_tree_mode_default",  "gop_show_times_default",  "gog_expiration_default",
  "gog_goptree_default",  "gog_show_times_default",
};
#endif /* NDEBUG */

#ifndef NDEBUG
/* For tracing reduce actions, the names of all rules are required.
*/
static const char *const yyRuleName[] = {
 /*   0 */ "mate_config ::= decls",
 /*   1 */ "decls ::= decls decl",
 /*   2 */ "decls ::=",
 /*   3 */ "decl ::= pdu_decl",
 /*   4 */ "decl ::= gop_decl",
 /*   5 */ "decl ::= gog_decl",
 /*   6 */ "decl ::= transform_decl",
 /*   7 */ "decl ::= defaults_decl",
 /*   8 */ "decl ::= debug_decl",
 /*   9 */ "decl ::= DONE_KW SEMICOLON",
 /*  10 */ "debug_decl ::= DEBUG_KW OPEN_BRACE dbgfile_default dbglevel_default pdu_dbglevel_default gop_dbglevel_default gog_dbglevel_default CLOSE_BRACE SEMICOLON",
 /*  11 */ "dbgfile_default ::= FILENAME_KW QUOTED SEMICOLON",
 /*  12 */ "dbgfile_default ::= FILENAME_KW NAME SEMICOLON",
 /*  13 */ "dbgfile_default ::=",
 /*  14 */ "dbglevel_default ::= LEVEL_KW INTEGER SEMICOLON",
 /*  15 */ "dbglevel_default ::=",
 /*  16 */ "pdu_dbglevel_default ::= PDU_KW LEVEL_KW INTEGER SEMICOLON",
 /*  17 */ "pdu_dbglevel_default ::=",
 /*  18 */ "gop_dbglevel_default ::= GOP_KW LEVEL_KW INTEGER SEMICOLON",
 /*  19 */ "gop_dbglevel_default ::=",
 /*  20 */ "gog_dbglevel_default ::= GOG_KW LEVEL_KW INTEGER SEMICOLON",
 /*  21 */ "gog_dbglevel_default ::=",
 /*  22 */ "defaults_decl ::= DEFAULT_KW OPEN_BRACE pdu_defaults gop_defaults gog_defaults CLOSE_BRACE SEMICOLON",
 /*  23 */ "pdu_defaults ::= PDU_KW OPEN_BRACE pdu_last_extracted_default pdu_drop_unassigned_default pdu_discard_default CLOSE_BRACE SEMICOLON",
 /*  24 */ "pdu_defaults ::=",
 /*  25 */ "pdu_last_extracted_default ::= LAST_EXTRACTED_KW true_false SEMICOLON",
 /*  26 */ "pdu_last_extracted_default ::=",
 /*  27 */ "pdu_drop_unassigned_default ::= DROP_UNASSIGNED_KW true_false SEMICOLON",
 /*  28 */ "pdu_drop_unassigned_default ::=",
 /*  29 */ "pdu_discard_default ::= DISCARD_PDU_DATA_KW true_false SEMICOLON",
 /*  30 */ "pdu_discard_default ::=",
 /*  31 */ "gop_defaults ::= GOP_KW OPEN_BRACE gop_expiration_default gop_idle_timeout_default gop_lifetime_default gop_drop_unassigned_default gop_tree_mode_default gop_show_times_default CLOSE_BRACE SEMICOLON",
 /*  32 */ "gop_defaults ::=",
 /*  33 */ "gop_expiration_default ::= EXPIRATION_KW time_value SEMICOLON",
 /*  34 */ "gop_expiration_default ::=",
 /*  35 */ "gop_idle_timeout_default ::= IDLE_TIMEOUT_KW time_value SEMICOLON",
 /*  36 */ "gop_idle_timeout_default ::=",
 /*  37 */ "gop_lifetime_default ::= LIFETIME_KW time_value SEMICOLON",
 /*  38 */ "gop_lifetime_default ::=",
 /*  39 */ "gop_drop_unassigned_default ::= DROP_UNASSIGNED_KW true_false SEMICOLON",
 /*  40 */ "gop_drop_unassigned_default ::=",
 /*  41 */ "gop_tree_mode_default ::= SHOW_TREE_KW gop_tree_mode SEMICOLON",
 /*  42 */ "gop_tree_mode_default ::=",
 /*  43 */ "gop_show_times_default ::= SHOW_TIMES_KW true_false SEMICOLON",
 /*  44 */ "gop_show_times_default ::=",
 /*  45 */ "gog_defaults ::= GOG_KW OPEN_BRACE gog_expiration_default gop_tree_mode_default gog_goptree_default gog_show_times_default CLOSE_BRACE SEMICOLON",
 /*  46 */ "gog_defaults ::=",
 /*  47 */ "gog_expiration_default ::= EXPIRATION_KW time_value SEMICOLON",
 /*  48 */ "gog_expiration_default ::=",
 /*  49 */ "gog_goptree_default ::= GOP_TREE_KW gop_tree_type SEMICOLON",
 /*  50 */ "gog_goptree_default ::=",
 /*  51 */ "gog_show_times_default ::= SHOW_TIMES_KW true_false SEMICOLON",
 /*  52 */ "gog_show_times_default ::=",
 /*  53 */ "transform_decl ::= TRANSFORM_KW NAME transform_body SEMICOLON",
 /*  54 */ "transform_body ::= OPEN_BRACE transform_statements CLOSE_BRACE",
 /*  55 */ "transform_statements ::= transform_statements transform_statement",
 /*  56 */ "transform_statements ::= transform_statement",
 /*  57 */ "transform_statement ::= transform_match transform_action SEMICOLON",
 /*  58 */ "transform_match ::= MATCH_KW match_mode avpl",
 /*  59 */ "transform_match ::=",
 /*  60 */ "transform_action ::=",
 /*  61 */ "transform_action ::= action_mode avpl",
 /*  62 */ "match_mode ::=",
 /*  63 */ "match_mode ::= STRICT_KW",
 /*  64 */ "match_mode ::= EVERY_KW",
 /*  65 */ "match_mode ::= LOOSE_KW",
 /*  66 */ "action_mode ::= REPLACE_KW",
 /*  67 */ "action_mode ::= INSERT_KW",
 /*  68 */ "action_mode ::=",
 /*  69 */ "pdu_decl ::= PDU_KW NAME PROTO_KW field TRANSPORT_KW proto_stack OPEN_BRACE payload_statement extraction_statements transform_list_statement criteria_statement pdu_drop_unassigned_statement discard_pdu_data_statement last_extracted_statement CLOSE_BRACE SEMICOLON",
 /*  70 */ "payload_statement ::=",
 /*  71 */ "payload_statement ::= PAYLOAD_KW proto_stack SEMICOLON",
 /*  72 */ "criteria_statement ::=",
 /*  73 */ "criteria_statement ::= CRITERIA_KW accept_mode match_mode avpl SEMICOLON",
 /*  74 */ "accept_mode ::=",
 /*  75 */ "accept_mode ::= ACCEPT_KW",
 /*  76 */ "accept_mode ::= REJECT_KW",
 /*  77 */ "extraction_statements ::= extraction_statements extraction_statement",
 /*  78 */ "extraction_statements ::= extraction_statement",
 /*  79 */ "extraction_statement ::= EXTRACT_KW NAME FROM_KW field SEMICOLON",
 /*  80 */ "pdu_drop_unassigned_statement ::= DROP_UNASSIGNED_KW true_false SEMICOLON",
 /*  81 */ "pdu_drop_unassigned_statement ::=",
 /*  82 */ "discard_pdu_data_statement ::= DISCARD_PDU_DATA_KW true_false SEMICOLON",
 /*  83 */ "discard_pdu_data_statement ::=",
 /*  84 */ "last_extracted_statement ::= LAST_PDU_KW true_false SEMICOLON",
 /*  85 */ "last_extracted_statement ::=",
 /*  86 */ "proto_stack ::= proto_stack SLASH field",
 /*  87 */ "proto_stack ::= field",
 /*  88 */ "field ::= NAME",
 /*  89 */ "gop_decl ::= GOP_KW NAME ON_KW pdu_name MATCH_KW avpl OPEN_BRACE gop_start_statement gop_stop_statement extra_statement transform_list_statement gop_expiration_statement idle_timeout_statement lifetime_statement gop_drop_unassigned_statement show_goptree_statement show_times_statement CLOSE_BRACE SEMICOLON",
 /*  90 */ "gop_drop_unassigned_statement ::= DROP_UNASSIGNED_KW true_false SEMICOLON",
 /*  91 */ "gop_drop_unassigned_statement ::=",
 /*  92 */ "gop_start_statement ::= START_KW avpl SEMICOLON",
 /*  93 */ "gop_start_statement ::=",
 /*  94 */ "gop_stop_statement ::= STOP_KW avpl SEMICOLON",
 /*  95 */ "gop_stop_statement ::=",
 /*  96 */ "show_goptree_statement ::= SHOW_TREE_KW gop_tree_mode SEMICOLON",
 /*  97 */ "show_goptree_statement ::=",
 /*  98 */ "show_times_statement ::= SHOW_TIMES_KW true_false SEMICOLON",
 /*  99 */ "show_times_statement ::=",
 /* 100 */ "gop_expiration_statement ::= EXPIRATION_KW time_value SEMICOLON",
 /* 101 */ "gop_expiration_statement ::=",
 /* 102 */ "idle_timeout_statement ::= IDLE_TIMEOUT_KW time_value SEMICOLON",
 /* 103 */ "idle_timeout_statement ::=",
 /* 104 */ "lifetime_statement ::= LIFETIME_KW time_value SEMICOLON",
 /* 105 */ "lifetime_statement ::=",
 /* 106 */ "gop_tree_mode ::= NO_TREE_KW",
 /* 107 */ "gop_tree_mode ::= PDU_TREE_KW",
 /* 108 */ "gop_tree_mode ::= FRAME_TREE_KW",
 /* 109 */ "gop_tree_mode ::= BASIC_TREE_KW",
 /* 110 */ "true_false ::= TRUE_KW",
 /* 111 */ "true_false ::= FALSE_KW",
 /* 112 */ "pdu_name ::= NAME",
 /* 113 */ "time_value ::= FLOATING",
 /* 114 */ "time_value ::= INTEGER",
 /* 115 */ "gog_decl ::= GOG_KW NAME OPEN_BRACE gog_key_statements extra_statement transform_list_statement gog_expiration_statement gog_goptree_statement show_times_statement CLOSE_BRACE SEMICOLON",
 /* 116 */ "gog_goptree_statement ::= GOP_TREE_KW gop_tree_type SEMICOLON",
 /* 117 */ "gog_goptree_statement ::=",
 /* 118 */ "gog_expiration_statement ::= EXPIRATION_KW time_value SEMICOLON",
 /* 119 */ "gog_expiration_statement ::=",
 /* 120 */ "gop_tree_type ::= NULL_TREE_KW",
 /* 121 */ "gop_tree_type ::= FULL_TREE_KW",
 /* 122 */ "gop_tree_type ::= BASIC_TREE_KW",
 /* 123 */ "gog_key_statements ::= gog_key_statements gog_key_statement",
 /* 124 */ "gog_key_statements ::= gog_key_statement",
 /* 125 */ "gog_key_statement ::= MEMBER_KW gop_name avpl SEMICOLON",
 /* 126 */ "gop_name ::= NAME",
 /* 127 */ "extra_statement ::= EXTRA_KW avpl SEMICOLON",
 /* 128 */ "extra_statement ::=",
 /* 129 */ "transform_list_statement ::= TRANSFORM_KW transform_list SEMICOLON",
 /* 130 */ "transform_list_statement ::=",
 /* 131 */ "transform_list ::= transform_list COMMA transform",
 /* 132 */ "transform_list ::= transform",
 /* 133 */ "transform ::= NAME",
 /* 134 */ "avpl ::= OPEN_PARENS avps CLOSE_PARENS",
 /* 135 */ "avpl ::= OPEN_PARENS CLOSE_PARENS",
 /* 136 */ "avps ::= avps COMMA avp",
 /* 137 */ "avps ::= avp",
 /* 138 */ "avp ::= NAME AVP_OPERATOR value",
 /* 139 */ "avp ::= NAME",
 /* 140 */ "avp ::= NAME OPEN_BRACE avp_oneoff CLOSE_BRACE",
 /* 141 */ "avp_oneoff ::= avp_oneoff PIPE value",
 /* 142 */ "avp_oneoff ::= value",
 /* 143 */ "value ::= QUOTED",
 /* 144 */ "value ::= NAME",
 /* 145 */ "value ::= FLOATING",
 /* 146 */ "value ::= INTEGER",
 /* 147 */ "value ::= DOTED_IP",
 /* 148 */ "value ::= COLONIZED",
};
#endif /* NDEBUG */


#if YYSTACKDEPTH<=0
/*
** Try to increase the size of the parser stack.
*/
static void yyGrowStack(yyParser *p){
  int newSize;
  yyStackEntry *pNew;

  newSize = p->yystksz*2 + 100;
  pNew = realloc(p->yystack, newSize*sizeof(pNew[0]));
  if( pNew ){
    p->yystack = pNew;
    p->yystksz = newSize;
#ifndef NDEBUG
    if( yyTraceFILE ){
      fprintf(yyTraceFILE,"%sStack grows to %d entries!\n",
              yyTracePrompt, p->yystksz);
    }
#endif
  }
}
#endif

/* Datatype of the argument to the memory allocated passed as the
** second argument to MateParserAlloc() below.  This can be changed by
** putting an appropriate #define in the %include section of the input
** grammar.
*/
#ifndef YYMALLOCARGTYPE
# define YYMALLOCARGTYPE size_t
#endif

/*
** This function allocates a new parser.
** The only argument is a pointer to a function which works like
** malloc.
**
** Inputs:
** A pointer to the function used to allocate memory.
**
** Outputs:
** A pointer to a parser.  This pointer is used in subsequent calls
** to MateParser and MateParserFree.
*/
void *MateParserAlloc(void *(*mallocProc)(YYMALLOCARGTYPE)){
  yyParser *pParser;
  pParser = (yyParser*)(*mallocProc)( (YYMALLOCARGTYPE)sizeof(yyParser) );
  if( pParser ){
    pParser->yyidx = -1;
#ifdef YYTRACKMAXSTACKDEPTH
    pParser->yyidxMax = 0;
#endif
#if YYSTACKDEPTH<=0
    pParser->yystack = NULL;
    pParser->yystksz = 0;
    yyGrowStack(pParser);
#endif
  }
  return pParser;
}

/* The following function deletes the "minor type" or semantic value
** associated with a symbol.  The symbol can be either a terminal
** or nonterminal. "yymajor" is the symbol code, and "yypminor" is
** a pointer to the value to be deleted.  The code used to do the
** deletions is derived from the %destructor and/or %token_destructor
** directives of the input grammar.
*/
static void yy_destructor(
  yyParser *yypParser,    /* The parser */
  YYCODETYPE yymajor,     /* Type code for object to destroy */
  YYMINORTYPE *yypminor   /* The object to be destroyed */
){
  MateParserARG_FETCH;
  switch( yymajor ){
    /* Here is inserted the actions which take place when a
    ** terminal or non-terminal is destroyed.  This can happen
    ** when the symbol is popped from the stack during a
    ** reduce or during error processing or when a parser is
    ** being destroyed before it is finished parsing.
    **
    ** Note: during a reduce, the only symbols destroyed are those
    ** which appear on the RHS of the rule, but which are *not* used
    ** inside the C code.
    */
/********* Begin destructor definitions ***************************************/
      /* TERMINAL Destructor */
    case 1: /* DONE_KW */
    case 2: /* SEMICOLON */
    case 3: /* DEBUG_KW */
    case 4: /* OPEN_BRACE */
    case 5: /* CLOSE_BRACE */
    case 6: /* FILENAME_KW */
    case 7: /* QUOTED */
    case 8: /* NAME */
    case 9: /* LEVEL_KW */
    case 10: /* INTEGER */
    case 11: /* PDU_KW */
    case 12: /* GOP_KW */
    case 13: /* GOG_KW */
    case 14: /* DEFAULT_KW */
    case 15: /* LAST_EXTRACTED_KW */
    case 16: /* DROP_UNASSIGNED_KW */
    case 17: /* DISCARD_PDU_DATA_KW */
    case 18: /* EXPIRATION_KW */
    case 19: /* IDLE_TIMEOUT_KW */
    case 20: /* LIFETIME_KW */
    case 21: /* SHOW_TREE_KW */
    case 22: /* SHOW_TIMES_KW */
    case 23: /* GOP_TREE_KW */
    case 24: /* TRANSFORM_KW */
    case 25: /* MATCH_KW */
    case 26: /* STRICT_KW */
    case 27: /* EVERY_KW */
    case 28: /* LOOSE_KW */
    case 29: /* REPLACE_KW */
    case 30: /* INSERT_KW */
    case 31: /* PROTO_KW */
    case 32: /* TRANSPORT_KW */
    case 33: /* PAYLOAD_KW */
    case 34: /* CRITERIA_KW */
    case 35: /* ACCEPT_KW */
    case 36: /* REJECT_KW */
    case 37: /* EXTRACT_KW */
    case 38: /* FROM_KW */
    case 39: /* LAST_PDU_KW */
    case 40: /* SLASH */
    case 41: /* ON_KW */
    case 42: /* START_KW */
    case 43: /* STOP_KW */
    case 44: /* NO_TREE_KW */
    case 45: /* PDU_TREE_KW */
    case 46: /* FRAME_TREE_KW */
    case 47: /* BASIC_TREE_KW */
    case 48: /* TRUE_KW */
    case 49: /* FALSE_KW */
    case 50: /* FLOATING */
    case 51: /* NULL_TREE_KW */
    case 52: /* FULL_TREE_KW */
    case 53: /* MEMBER_KW */
    case 54: /* EXTRA_KW */
    case 55: /* COMMA */
    case 56: /* OPEN_PARENS */
    case 57: /* CLOSE_PARENS */
    case 58: /* AVP_OPERATOR */
    case 59: /* PIPE */
    case 60: /* DOTED_IP */
    case 61: /* COLONIZED */
{
#line 179 "./mate_grammar.lemon"

	(void) mc; /* Mark unused, similar to Q_UNUSED */
	if ((yypminor->yy0)) g_free((yypminor->yy0));

#line 952 "mate_grammar.c"
}
      break;
/********* End destructor definitions *****************************************/
    default:  break;   /* If no destructor action specified: do nothing */
  }
}

/*
** Pop the parser's stack once.
**
** If there is a destructor routine associated with the token which
** is popped from the stack, then call it.
*/
static void yy_pop_parser_stack(yyParser *pParser){
  yyStackEntry *yytos;
  assert( pParser->yyidx>=0 );
  yytos = &pParser->yystack[pParser->yyidx--];
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sPopping %s\n",
      yyTracePrompt,
      yyTokenName[yytos->major]);
  }
#endif
  yy_destructor(pParser, yytos->major, &yytos->minor);
}

/*
** Deallocate and destroy a parser.  Destructors are called for
** all stack elements before shutting the parser down.
**
** If the YYPARSEFREENEVERNULL macro exists (for example because it
** is defined in a %include section of the input grammar) then it is
** assumed that the input pointer is never NULL.
*/
void MateParserFree(
  void *p,                    /* The parser to be deleted */
  void (*freeProc)(void*)     /* Function used to reclaim memory */
){
  yyParser *pParser = (yyParser*)p;
#ifndef YYPARSEFREENEVERNULL
  if( pParser==0 ) return;
#endif
  while( pParser->yyidx>=0 ) yy_pop_parser_stack(pParser);
#if YYSTACKDEPTH<=0
  free(pParser->yystack);
#endif
  (*freeProc)((void*)pParser);
}

/*
** Return the peak depth of the stack for a parser.
*/
#ifdef YYTRACKMAXSTACKDEPTH
int MateParserStackPeak(void *p){
  yyParser *pParser = (yyParser*)p;
  return pParser->yyidxMax;
}
#endif

/*
** Find the appropriate action for a parser given the terminal
** look-ahead token iLookAhead.
*/
static int yy_find_shift_action(
  yyParser *pParser,        /* The parser */
  YYCODETYPE iLookAhead     /* The look-ahead token */
){
  int i;
  int stateno = pParser->yystack[pParser->yyidx].stateno;

  if( stateno>=YY_MIN_REDUCE ) return stateno;
  assert( stateno <= YY_SHIFT_COUNT );
  do{
    i = yy_shift_ofst[stateno];
    if( i==YY_SHIFT_USE_DFLT ) return yy_default[stateno];
    assert( iLookAhead!=YYNOCODE );
    i += iLookAhead;
    if( i<0 || i>=YY_ACTTAB_COUNT || yy_lookahead[i]!=iLookAhead ){
      if( iLookAhead>0 ){
#ifdef YYFALLBACK
        YYCODETYPE iFallback;            /* Fallback token */
        if( iLookAhead<sizeof(yyFallback)/sizeof(yyFallback[0])
               && (iFallback = yyFallback[iLookAhead])!=0 ){
#ifndef NDEBUG
          if( yyTraceFILE ){
            fprintf(yyTraceFILE, "%sFALLBACK %s => %s\n",
               yyTracePrompt, yyTokenName[iLookAhead], yyTokenName[iFallback]);
          }
#endif
          assert( yyFallback[iFallback]==0 ); /* Fallback loop must terminate */
          iLookAhead = iFallback;
          continue;
        }
#endif
#ifdef YYWILDCARD
        {
          int j = i - iLookAhead + YYWILDCARD;
          if(
#if YY_SHIFT_MIN+YYWILDCARD<0
            j>=0 &&
#endif
#if YY_SHIFT_MAX+YYWILDCARD>=YY_ACTTAB_COUNT
            j<YY_ACTTAB_COUNT &&
#endif
            yy_lookahead[j]==YYWILDCARD
          ){
#ifndef NDEBUG
            if( yyTraceFILE ){
              fprintf(yyTraceFILE, "%sWILDCARD %s => %s\n",
                 yyTracePrompt, yyTokenName[iLookAhead],
                 yyTokenName[YYWILDCARD]);
            }
#endif /* NDEBUG */
            return yy_action[j];
          }
        }
#endif /* YYWILDCARD */
      }
      return yy_default[stateno];
    }else{
      return yy_action[i];
    }
  }while(1);
}

/*
** Find the appropriate action for a parser given the non-terminal
** look-ahead token iLookAhead.
*/
static int yy_find_reduce_action(
  int stateno,              /* Current state number */
  YYCODETYPE iLookAhead     /* The look-ahead token */
){
  int i;
#ifdef YYERRORSYMBOL
  if( stateno>YY_REDUCE_COUNT ){
    return yy_default[stateno];
  }
#else
  assert( stateno<=YY_REDUCE_COUNT );
#endif
  i = yy_reduce_ofst[stateno];
  assert( i!=YY_REDUCE_USE_DFLT );
  assert( iLookAhead!=YYNOCODE );
  i += iLookAhead;
#ifdef YYERRORSYMBOL
  if( i<0 || i>=YY_ACTTAB_COUNT || yy_lookahead[i]!=iLookAhead ){
    return yy_default[stateno];
  }
#else
  assert( i>=0 && i<YY_ACTTAB_COUNT );
  assert( yy_lookahead[i]==iLookAhead );
#endif
  return yy_action[i];
}

/*
** The following routine is called if the stack overflows.
*/
static void yyStackOverflow(yyParser *yypParser, YYMINORTYPE *yypMinor _U_){
   MateParserARG_FETCH;
   yypParser->yyidx--;
#ifndef NDEBUG
   if( yyTraceFILE ){
     fprintf(yyTraceFILE,"%sStack Overflow!\n",yyTracePrompt);
   }
#endif
   while( yypParser->yyidx>=0 ) yy_pop_parser_stack(yypParser);
   /* Here code is inserted which will execute if the parser
   ** stack every overflows */
/******** Begin %stack_overflow code ******************************************/
/******** End %stack_overflow code ********************************************/
   MateParserARG_STORE; /* Suppress warning about unused %extra_argument var */
}

/*
** Print tracing information for a SHIFT action
*/
#ifndef NDEBUG
static void yyTraceShift(yyParser *yypParser, int yyNewState){
  if( yyTraceFILE ){
    if( yyNewState<YYNSTATE ){
      fprintf(yyTraceFILE,"%sShift '%s', go to state %d\n",
         yyTracePrompt,yyTokenName[yypParser->yystack[yypParser->yyidx].major],
         yyNewState);
    }else{
      fprintf(yyTraceFILE,"%sShift '%s'\n",
         yyTracePrompt,yyTokenName[yypParser->yystack[yypParser->yyidx].major]);
    }
  }
}
#else
# define yyTraceShift(X,Y)
#endif

/*
** Perform a shift action.
*/
static void yy_shift(
  yyParser *yypParser,          /* The parser to be shifted */
  int yyNewState,               /* The new state to shift in */
  int yyMajor,                  /* The major token to shift in */
  YYMINORTYPE *yypMinor         /* Pointer to the minor token to shift in */
){
  yyStackEntry *yytos;
  yypParser->yyidx++;
#ifdef YYTRACKMAXSTACKDEPTH
  if( yypParser->yyidx>yypParser->yyidxMax ){
    yypParser->yyidxMax = yypParser->yyidx;
  }
#endif
#if YYSTACKDEPTH>0
  if( yypParser->yyidx>=YYSTACKDEPTH ){
    yyStackOverflow(yypParser, yypMinor);
    return;
  }
#else
  if( yypParser->yyidx>=yypParser->yystksz ){
    yyGrowStack(yypParser);
    if( yypParser->yyidx>=yypParser->yystksz ){
      yyStackOverflow(yypParser, yypMinor);
      return;
    }
  }
#endif
  yytos = &yypParser->yystack[yypParser->yyidx];
  yytos->stateno = (YYACTIONTYPE)yyNewState;
  yytos->major = (YYCODETYPE)yyMajor;
  yytos->minor = *yypMinor;
  yyTraceShift(yypParser, yyNewState);
}

/* The following table contains information about every rule that
** is used during the reduce.
*/
static const struct {
  YYCODETYPE lhs;         /* Symbol on the left-hand side of the rule */
  unsigned char nrhs;     /* Number of right-hand side symbols in the rule */
} yyRuleInfo[] = {
  { 110, 1 },
  { 111, 2 },
  { 111, 0 },
  { 112, 1 },
  { 112, 1 },
  { 112, 1 },
  { 112, 1 },
  { 112, 1 },
  { 112, 1 },
  { 112, 2 },
  { 117, 9 },
  { 118, 3 },
  { 118, 3 },
  { 118, 0 },
  { 119, 3 },
  { 119, 0 },
  { 120, 4 },
  { 120, 0 },
  { 121, 4 },
  { 121, 0 },
  { 122, 4 },
  { 122, 0 },
  { 116, 7 },
  { 123, 7 },
  { 123, 0 },
  { 126, 3 },
  { 126, 0 },
  { 127, 3 },
  { 127, 0 },
  { 128, 3 },
  { 128, 0 },
  { 124, 10 },
  { 124, 0 },
  { 129, 3 },
  { 129, 0 },
  { 130, 3 },
  { 130, 0 },
  { 131, 3 },
  { 131, 0 },
  { 132, 3 },
  { 132, 0 },
  { 133, 3 },
  { 133, 0 },
  { 134, 3 },
  { 134, 0 },
  { 125, 8 },
  { 125, 0 },
  { 135, 3 },
  { 135, 0 },
  { 136, 3 },
  { 136, 0 },
  { 137, 3 },
  { 137, 0 },
  { 63, 4 },
  { 64, 3 },
  { 65, 2 },
  { 65, 1 },
  { 66, 3 },
  { 67, 3 },
  { 67, 0 },
  { 68, 0 },
  { 68, 2 },
  { 69, 0 },
  { 69, 1 },
  { 69, 1 },
  { 69, 1 },
  { 70, 1 },
  { 70, 1 },
  { 70, 0 },
  { 113, 16 },
  { 101, 0 },
  { 101, 3 },
  { 76, 0 },
  { 76, 5 },
  { 77, 0 },
  { 77, 1 },
  { 77, 1 },
  { 82, 2 },
  { 82, 1 },
  { 81, 5 },
  { 78, 3 },
  { 78, 0 },
  { 79, 3 },
  { 79, 0 },
  { 80, 3 },
  { 80, 0 },
  { 102, 3 },
  { 102, 1 },
  { 103, 1 },
  { 114, 19 },
  { 87, 3 },
  { 87, 0 },
  { 84, 3 },
  { 84, 0 },
  { 85, 3 },
  { 85, 0 },
  { 88, 3 },
  { 88, 0 },
  { 89, 3 },
  { 89, 0 },
  { 90, 3 },
  { 90, 0 },
  { 91, 3 },
  { 91, 0 },
  { 92, 3 },
  { 92, 0 },
  { 74, 1 },
  { 74, 1 },
  { 74, 1 },
  { 74, 1 },
  { 75, 1 },
  { 75, 1 },
  { 73, 1 },
  { 72, 1 },
  { 72, 1 },
  { 115, 11 },
  { 95, 3 },
  { 95, 0 },
  { 94, 3 },
  { 94, 0 },
  { 100, 1 },
  { 100, 1 },
  { 100, 1 },
  { 96, 2 },
  { 96, 1 },
  { 97, 4 },
  { 71, 1 },
  { 86, 3 },
  { 86, 0 },
  { 98, 3 },
  { 98, 0 },
  { 104, 3 },
  { 104, 1 },
  { 99, 1 },
  { 105, 3 },
  { 105, 2 },
  { 106, 3 },
  { 106, 1 },
  { 107, 3 },
  { 107, 1 },
  { 107, 4 },
  { 109, 3 },
  { 109, 1 },
  { 108, 1 },
  { 108, 1 },
  { 108, 1 },
  { 108, 1 },
  { 108, 1 },
  { 108, 1 },
};

static void yy_accept(yyParser*);  /* Forward Declaration */

/*
** Perform a reduce action and the shift that must immediately
** follow the reduce.
*/
static void yy_reduce(
  yyParser *yypParser,         /* The parser */
  int yyruleno                 /* Number of the rule by which to reduce */
){
  int yygoto;                     /* The next state */
  int yyact;                      /* The next action */
  YYMINORTYPE yygotominor;        /* The LHS of the rule reduced */
  yyStackEntry *yymsp;            /* The top of the parser's stack */
  int yysize;                     /* Amount to pop the stack */
  MateParserARG_FETCH;
  yymsp = &yypParser->yystack[yypParser->yyidx];
#ifndef NDEBUG
  if( yyTraceFILE && yyruleno>=0
        && yyruleno<(int)(sizeof(yyRuleName)/sizeof(yyRuleName[0])) ){
    yysize = yyRuleInfo[yyruleno].nrhs;
    fprintf(yyTraceFILE, "%sReduce [%s], go to state %d.\n", yyTracePrompt,
      yyRuleName[yyruleno], yymsp[-yysize].stateno);
  }
#endif /* NDEBUG */
  yygotominor = yyzerominor;

  switch( yyruleno ){
  /* Beginning here are the reduction cases.  A typical example
  ** follows:
  **   case 0:
  **  #line <lineno> <grammarfile>
  **     { ... }           // User supplied code
  **  #line <lineno> <thisfile>
  **     break;
  */
/********** Begin reduce actions **********************************************/
      case 9: /* decl ::= DONE_KW SEMICOLON */
#line 261 "mate_grammar.lemon"
{
  yy_destructor(yypParser,1,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1387 "mate_grammar.c"
        break;
      case 10: /* debug_decl ::= DEBUG_KW OPEN_BRACE dbgfile_default dbglevel_default pdu_dbglevel_default gop_dbglevel_default gog_dbglevel_default CLOSE_BRACE SEMICOLON */
#line 266 "mate_grammar.lemon"
{
  yy_destructor(yypParser,3,&yymsp[-8].minor);
  yy_destructor(yypParser,4,&yymsp[-7].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1397 "mate_grammar.c"
        break;
      case 11: /* dbgfile_default ::= FILENAME_KW QUOTED SEMICOLON */
#line 268 "mate_grammar.lemon"
{ mc->dbg_facility = ws_fopen(yymsp[-1].minor.yy0,"w"); if (mc->dbg_facility == NULL) report_open_failure(yymsp[-1].minor.yy0,errno,TRUE);   yy_destructor(yypParser,6,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1404 "mate_grammar.c"
        break;
      case 12: /* dbgfile_default ::= FILENAME_KW NAME SEMICOLON */
#line 269 "mate_grammar.lemon"
{ mc->dbg_facility = ws_fopen(yymsp[-1].minor.yy0,"w"); if (mc->dbg_facility == NULL) report_open_failure(yymsp[-1].minor.yy0,errno,TRUE);    yy_destructor(yypParser,6,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1411 "mate_grammar.c"
        break;
      case 14: /* dbglevel_default ::= LEVEL_KW INTEGER SEMICOLON */
#line 272 "mate_grammar.lemon"
{ mc->dbg_lvl = (int) strtol(yymsp[-1].minor.yy0,NULL,10);   yy_destructor(yypParser,9,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1418 "mate_grammar.c"
        break;
      case 16: /* pdu_dbglevel_default ::= PDU_KW LEVEL_KW INTEGER SEMICOLON */
#line 275 "mate_grammar.lemon"
{ mc->dbg_pdu_lvl = (int) strtol(yymsp[-1].minor.yy0,NULL,10);   yy_destructor(yypParser,11,&yymsp[-3].minor);
  yy_destructor(yypParser,9,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1426 "mate_grammar.c"
        break;
      case 18: /* gop_dbglevel_default ::= GOP_KW LEVEL_KW INTEGER SEMICOLON */
#line 278 "mate_grammar.lemon"
{ mc->dbg_gop_lvl = (int) strtol(yymsp[-1].minor.yy0,NULL,10);   yy_destructor(yypParser,12,&yymsp[-3].minor);
  yy_destructor(yypParser,9,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1434 "mate_grammar.c"
        break;
      case 20: /* gog_dbglevel_default ::= GOG_KW LEVEL_KW INTEGER SEMICOLON */
#line 281 "mate_grammar.lemon"
{ mc->dbg_gog_lvl = (int) strtol(yymsp[-1].minor.yy0,NULL,10);   yy_destructor(yypParser,13,&yymsp[-3].minor);
  yy_destructor(yypParser,9,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1442 "mate_grammar.c"
        break;
      case 22: /* defaults_decl ::= DEFAULT_KW OPEN_BRACE pdu_defaults gop_defaults gog_defaults CLOSE_BRACE SEMICOLON */
#line 288 "mate_grammar.lemon"
{
  yy_destructor(yypParser,14,&yymsp[-6].minor);
  yy_destructor(yypParser,4,&yymsp[-5].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1452 "mate_grammar.c"
        break;
      case 23: /* pdu_defaults ::= PDU_KW OPEN_BRACE pdu_last_extracted_default pdu_drop_unassigned_default pdu_discard_default CLOSE_BRACE SEMICOLON */
#line 290 "mate_grammar.lemon"
{
  yy_destructor(yypParser,11,&yymsp[-6].minor);
  yy_destructor(yypParser,4,&yymsp[-5].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1462 "mate_grammar.c"
        break;
      case 25: /* pdu_last_extracted_default ::= LAST_EXTRACTED_KW true_false SEMICOLON */
#line 293 "mate_grammar.lemon"
{ mc->defaults.pdu.last_extracted = yymsp[-1].minor.yy216;   yy_destructor(yypParser,15,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1469 "mate_grammar.c"
        break;
      case 27: /* pdu_drop_unassigned_default ::= DROP_UNASSIGNED_KW true_false SEMICOLON */
#line 296 "mate_grammar.lemon"
{ mc->defaults.pdu.drop_unassigned = yymsp[-1].minor.yy216;   yy_destructor(yypParser,16,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1476 "mate_grammar.c"
        break;
      case 29: /* pdu_discard_default ::= DISCARD_PDU_DATA_KW true_false SEMICOLON */
#line 299 "mate_grammar.lemon"
{ mc->defaults.pdu.discard = yymsp[-1].minor.yy216;   yy_destructor(yypParser,17,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1483 "mate_grammar.c"
        break;
      case 31: /* gop_defaults ::= GOP_KW OPEN_BRACE gop_expiration_default gop_idle_timeout_default gop_lifetime_default gop_drop_unassigned_default gop_tree_mode_default gop_show_times_default CLOSE_BRACE SEMICOLON */
#line 302 "mate_grammar.lemon"
{
  yy_destructor(yypParser,12,&yymsp[-9].minor);
  yy_destructor(yypParser,4,&yymsp[-8].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1493 "mate_grammar.c"
        break;
      case 33: /* gop_expiration_default ::= EXPIRATION_KW time_value SEMICOLON */
      case 47: /* gog_expiration_default ::= EXPIRATION_KW time_value SEMICOLON */ yytestcase(yyruleno==47);
#line 305 "mate_grammar.lemon"
{ mc->defaults.gop.expiration = yymsp[-1].minor.yy83;   yy_destructor(yypParser,18,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1501 "mate_grammar.c"
        break;
      case 35: /* gop_idle_timeout_default ::= IDLE_TIMEOUT_KW time_value SEMICOLON */
#line 308 "mate_grammar.lemon"
{ mc->defaults.gop.idle_timeout = yymsp[-1].minor.yy83;   yy_destructor(yypParser,19,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1508 "mate_grammar.c"
        break;
      case 37: /* gop_lifetime_default ::= LIFETIME_KW time_value SEMICOLON */
#line 311 "mate_grammar.lemon"
{ mc->defaults.gop.lifetime = yymsp[-1].minor.yy83;   yy_destructor(yypParser,20,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1515 "mate_grammar.c"
        break;
      case 39: /* gop_drop_unassigned_default ::= DROP_UNASSIGNED_KW true_false SEMICOLON */
#line 314 "mate_grammar.lemon"
{ mc->defaults.gop.drop_unassigned = yymsp[-1].minor.yy216;   yy_destructor(yypParser,16,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1522 "mate_grammar.c"
        break;
      case 41: /* gop_tree_mode_default ::= SHOW_TREE_KW gop_tree_mode SEMICOLON */
#line 317 "mate_grammar.lemon"
{ mc->defaults.gop.pdu_tree_mode = (gop_pdu_tree_t)yymsp[-1].minor.yy93;   yy_destructor(yypParser,21,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1529 "mate_grammar.c"
        break;
      case 43: /* gop_show_times_default ::= SHOW_TIMES_KW true_false SEMICOLON */
#line 320 "mate_grammar.lemon"
{ mc->defaults.gop.show_times = yymsp[-1].minor.yy216;   yy_destructor(yypParser,22,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1536 "mate_grammar.c"
        break;
      case 45: /* gog_defaults ::= GOG_KW OPEN_BRACE gog_expiration_default gop_tree_mode_default gog_goptree_default gog_show_times_default CLOSE_BRACE SEMICOLON */
#line 323 "mate_grammar.lemon"
{
  yy_destructor(yypParser,13,&yymsp[-7].minor);
  yy_destructor(yypParser,4,&yymsp[-6].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1546 "mate_grammar.c"
        break;
      case 49: /* gog_goptree_default ::= GOP_TREE_KW gop_tree_type SEMICOLON */
#line 329 "mate_grammar.lemon"
{ mc->defaults.gog.gop_tree_mode = yymsp[-1].minor.yy93;   yy_destructor(yypParser,23,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1553 "mate_grammar.c"
        break;
      case 51: /* gog_show_times_default ::= SHOW_TIMES_KW true_false SEMICOLON */
#line 332 "mate_grammar.lemon"
{ mc->defaults.gog.show_times = yymsp[-1].minor.yy216;   yy_destructor(yypParser,22,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1560 "mate_grammar.c"
        break;
      case 53: /* transform_decl ::= TRANSFORM_KW NAME transform_body SEMICOLON */
#line 339 "mate_grammar.lemon"
{
	AVPL_Transf* c;

	if ( g_hash_table_lookup(mc->transfs,yymsp[-2].minor.yy0) ) {
		configuration_error(mc,"yygotominor.yy71 transformation called '%s' exists already",yymsp[-2].minor.yy0);
	}

	for ( c = yymsp[-1].minor.yy71; c; c = c->next )
		c->name = g_strdup(yymsp[-2].minor.yy0);

	g_hash_table_insert(mc->transfs,yymsp[-1].minor.yy71->name,yymsp[-1].minor.yy71);

	yygotominor.yy71 = NULL;
  yy_destructor(yypParser,24,&yymsp[-3].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1580 "mate_grammar.c"
        break;
      case 54: /* transform_body ::= OPEN_BRACE transform_statements CLOSE_BRACE */
#line 354 "mate_grammar.lemon"
{ yygotominor.yy71 = yymsp[-1].minor.yy71;   yy_destructor(yypParser,4,&yymsp[-2].minor);
  yy_destructor(yypParser,5,&yymsp[0].minor);
}
#line 1587 "mate_grammar.c"
        break;
      case 55: /* transform_statements ::= transform_statements transform_statement */
#line 356 "mate_grammar.lemon"
{
	AVPL_Transf* c;

	for ( c = yymsp[-1].minor.yy71; c->next; c = c->next ) ;
	c->next = yymsp[0].minor.yy71;
	yygotominor.yy71 = yymsp[-1].minor.yy71;
}
#line 1598 "mate_grammar.c"
        break;
      case 56: /* transform_statements ::= transform_statement */
#line 364 "mate_grammar.lemon"
{ yygotominor.yy71 = yymsp[0].minor.yy71; }
#line 1603 "mate_grammar.c"
        break;
      case 57: /* transform_statement ::= transform_match transform_action SEMICOLON */
#line 366 "mate_grammar.lemon"
{
	yygotominor.yy71 = new_transform_elem(yymsp[-2].minor.yy16->avpl,yymsp[-1].minor.yy77->avpl,yymsp[-2].minor.yy16->match_mode,yymsp[-1].minor.yy77->replace_mode);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1611 "mate_grammar.c"
        break;
      case 58: /* transform_match ::= MATCH_KW match_mode avpl */
#line 370 "mate_grammar.lemon"
{
	yygotominor.yy16 = (transf_match_t *)g_malloc(sizeof(transf_match_t));
	yygotominor.yy16->match_mode = yymsp[-1].minor.yy172;
	yygotominor.yy16->avpl = yymsp[0].minor.yy170;
  yy_destructor(yypParser,25,&yymsp[-2].minor);
}
#line 1621 "mate_grammar.c"
        break;
      case 59: /* transform_match ::= */
#line 376 "mate_grammar.lemon"
{
	yygotominor.yy16 = (transf_match_t *)g_malloc(sizeof(transf_match_t));
	yygotominor.yy16->match_mode = AVPL_STRICT;
	yygotominor.yy16->avpl = new_avpl("");

}
#line 1631 "mate_grammar.c"
        break;
      case 60: /* transform_action ::= */
#line 383 "mate_grammar.lemon"
{
	yygotominor.yy77 = (transf_action_t *)g_malloc(sizeof(transf_action_t));
	yygotominor.yy77->replace_mode = AVPL_INSERT;
	yygotominor.yy77->avpl = new_avpl("");
}
#line 1640 "mate_grammar.c"
        break;
      case 61: /* transform_action ::= action_mode avpl */
#line 388 "mate_grammar.lemon"
{
	yygotominor.yy77 = (transf_action_t *)g_malloc(sizeof(transf_action_t));
	yygotominor.yy77->replace_mode = yymsp[-1].minor.yy215;
	yygotominor.yy77->avpl = yymsp[0].minor.yy170;
}
#line 1649 "mate_grammar.c"
        break;
      case 62: /* match_mode ::= */
#line 394 "mate_grammar.lemon"
{ yygotominor.yy172 = AVPL_STRICT; }
#line 1654 "mate_grammar.c"
        break;
      case 63: /* match_mode ::= STRICT_KW */
#line 395 "mate_grammar.lemon"
{ yygotominor.yy172 = AVPL_STRICT;   yy_destructor(yypParser,26,&yymsp[0].minor);
}
#line 1660 "mate_grammar.c"
        break;
      case 64: /* match_mode ::= EVERY_KW */
#line 396 "mate_grammar.lemon"
{ yygotominor.yy172 = AVPL_EVERY;   yy_destructor(yypParser,27,&yymsp[0].minor);
}
#line 1666 "mate_grammar.c"
        break;
      case 65: /* match_mode ::= LOOSE_KW */
#line 397 "mate_grammar.lemon"
{ yygotominor.yy172 = AVPL_LOOSE;   yy_destructor(yypParser,28,&yymsp[0].minor);
}
#line 1672 "mate_grammar.c"
        break;
      case 66: /* action_mode ::= REPLACE_KW */
#line 399 "mate_grammar.lemon"
{ yygotominor.yy215 = AVPL_REPLACE;   yy_destructor(yypParser,29,&yymsp[0].minor);
}
#line 1678 "mate_grammar.c"
        break;
      case 67: /* action_mode ::= INSERT_KW */
#line 400 "mate_grammar.lemon"
{ yygotominor.yy215 = AVPL_INSERT;   yy_destructor(yypParser,30,&yymsp[0].minor);
}
#line 1684 "mate_grammar.c"
        break;
      case 68: /* action_mode ::= */
#line 401 "mate_grammar.lemon"
{ yygotominor.yy215 = AVPL_INSERT; }
#line 1689 "mate_grammar.c"
        break;
      case 69: /* pdu_decl ::= PDU_KW NAME PROTO_KW field TRANSPORT_KW proto_stack OPEN_BRACE payload_statement extraction_statements transform_list_statement criteria_statement pdu_drop_unassigned_statement discard_pdu_data_statement last_extracted_statement CLOSE_BRACE SEMICOLON */
#line 417 "mate_grammar.lemon"
{

	mate_cfg_pdu* cfg  = new_pducfg(yymsp[-14].minor.yy0);
	extraction_t *extraction, *next_extraction;
	GPtrArray* transport_stack = g_ptr_array_new();
	int i;

	if (! cfg ) configuration_error(mc,"could not create Pdu %s.",yymsp[-14].minor.yy0);

	cfg->hfid_proto = yymsp[-12].minor.yy202->id;

	cfg->last_extracted = yymsp[-2].minor.yy216;
	cfg->discard = yymsp[-3].minor.yy216;
	cfg->drop_unassigned = yymsp[-4].minor.yy216;

	g_string_append_printf(mc->protos_filter,"||%s",yymsp[-12].minor.yy202->abbrev);

	/* flip the transport_stack */
	for (i = yymsp[-10].minor.yy231->len - 1; yymsp[-10].minor.yy231->len; i--) {
		g_ptr_array_add(transport_stack,g_ptr_array_remove_index(yymsp[-10].minor.yy231,i));
	}

	g_ptr_array_free(yymsp[-10].minor.yy231,FALSE);

	cfg->transport_ranges = transport_stack;
	cfg->payload_ranges = yymsp[-8].minor.yy231;

	if (yymsp[-5].minor.yy41) {
		cfg->criterium = yymsp[-5].minor.yy41->criterium_avpl;
		cfg->criterium_match_mode = yymsp[-5].minor.yy41->criterium_match_mode;
		cfg->criterium_accept_mode = yymsp[-5].minor.yy41->criterium_accept_mode;
	}

	cfg->transforms = yymsp[-6].minor.yy231;

	for (extraction = yymsp[-7].minor.yy259; extraction; extraction = next_extraction) {
		next_extraction = extraction->next;

		if ( ! add_hfid(extraction->hfi, extraction->as, cfg->hfids_attr) ) {
			configuration_error(mc,"MATE: failed to create extraction rule '%s'",extraction->as);
		}

		g_free(extraction);
	}
  yy_destructor(yypParser,11,&yymsp[-15].minor);
  yy_destructor(yypParser,31,&yymsp[-13].minor);
  yy_destructor(yypParser,32,&yymsp[-11].minor);
  yy_destructor(yypParser,4,&yymsp[-9].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1744 "mate_grammar.c"
        break;
      case 70: /* payload_statement ::= */
#line 463 "mate_grammar.lemon"
{ yygotominor.yy231 = NULL; }
#line 1749 "mate_grammar.c"
        break;
      case 71: /* payload_statement ::= PAYLOAD_KW proto_stack SEMICOLON */
#line 464 "mate_grammar.lemon"
{ yygotominor.yy231 = yymsp[-1].minor.yy231;   yy_destructor(yypParser,33,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1756 "mate_grammar.c"
        break;
      case 72: /* criteria_statement ::= */
#line 466 "mate_grammar.lemon"
{ yygotominor.yy41 = NULL; }
#line 1761 "mate_grammar.c"
        break;
      case 73: /* criteria_statement ::= CRITERIA_KW accept_mode match_mode avpl SEMICOLON */
#line 467 "mate_grammar.lemon"
{
	yygotominor.yy41 = (pdu_criteria_t *)g_malloc(sizeof(pdu_criteria_t));
	yygotominor.yy41->criterium_avpl = yymsp[-1].minor.yy170;
	yygotominor.yy41->criterium_match_mode = yymsp[-2].minor.yy172;
	yygotominor.yy41->criterium_accept_mode = yymsp[-3].minor.yy188;
  yy_destructor(yypParser,34,&yymsp[-4].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1773 "mate_grammar.c"
        break;
      case 74: /* accept_mode ::= */
#line 474 "mate_grammar.lemon"
{ yygotominor.yy188 = ACCEPT_MODE; }
#line 1778 "mate_grammar.c"
        break;
      case 75: /* accept_mode ::= ACCEPT_KW */
#line 475 "mate_grammar.lemon"
{ yygotominor.yy188 = ACCEPT_MODE;   yy_destructor(yypParser,35,&yymsp[0].minor);
}
#line 1784 "mate_grammar.c"
        break;
      case 76: /* accept_mode ::= REJECT_KW */
#line 476 "mate_grammar.lemon"
{ yygotominor.yy188 = REJECT_MODE;   yy_destructor(yypParser,36,&yymsp[0].minor);
}
#line 1790 "mate_grammar.c"
        break;
      case 77: /* extraction_statements ::= extraction_statements extraction_statement */
#line 478 "mate_grammar.lemon"
{ yygotominor.yy259 = yymsp[-1].minor.yy259; yygotominor.yy259->last = yygotominor.yy259->last->next = yymsp[0].minor.yy259; }
#line 1795 "mate_grammar.c"
        break;
      case 78: /* extraction_statements ::= extraction_statement */
#line 479 "mate_grammar.lemon"
{ yygotominor.yy259 = yymsp[0].minor.yy259; yygotominor.yy259->last = yygotominor.yy259; }
#line 1800 "mate_grammar.c"
        break;
      case 79: /* extraction_statement ::= EXTRACT_KW NAME FROM_KW field SEMICOLON */
#line 481 "mate_grammar.lemon"
{
	yygotominor.yy259 = (extraction_t *)g_malloc(sizeof(extraction_t));
	yygotominor.yy259->as = yymsp[-3].minor.yy0;
	yygotominor.yy259->hfi = yymsp[-1].minor.yy202;
	yygotominor.yy259->next = yygotominor.yy259->last = NULL;
  yy_destructor(yypParser,37,&yymsp[-4].minor);
  yy_destructor(yypParser,38,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1813 "mate_grammar.c"
        break;
      case 80: /* pdu_drop_unassigned_statement ::= DROP_UNASSIGNED_KW true_false SEMICOLON */
      case 90: /* gop_drop_unassigned_statement ::= DROP_UNASSIGNED_KW true_false SEMICOLON */ yytestcase(yyruleno==90);
#line 489 "mate_grammar.lemon"
{ yygotominor.yy216 = yymsp[-1].minor.yy216;   yy_destructor(yypParser,16,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1821 "mate_grammar.c"
        break;
      case 81: /* pdu_drop_unassigned_statement ::= */
#line 490 "mate_grammar.lemon"
{ yygotominor.yy216 =  mc->defaults.pdu.drop_unassigned; }
#line 1826 "mate_grammar.c"
        break;
      case 82: /* discard_pdu_data_statement ::= DISCARD_PDU_DATA_KW true_false SEMICOLON */
#line 492 "mate_grammar.lemon"
{ yygotominor.yy216 = yymsp[-1].minor.yy216;   yy_destructor(yypParser,17,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1833 "mate_grammar.c"
        break;
      case 83: /* discard_pdu_data_statement ::= */
#line 493 "mate_grammar.lemon"
{ yygotominor.yy216 =  mc->defaults.pdu.discard; }
#line 1838 "mate_grammar.c"
        break;
      case 84: /* last_extracted_statement ::= LAST_PDU_KW true_false SEMICOLON */
#line 495 "mate_grammar.lemon"
{ yygotominor.yy216 = yymsp[-1].minor.yy216;   yy_destructor(yypParser,39,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1845 "mate_grammar.c"
        break;
      case 85: /* last_extracted_statement ::= */
#line 496 "mate_grammar.lemon"
{ yygotominor.yy216 = mc->defaults.pdu.last_extracted; }
#line 1850 "mate_grammar.c"
        break;
      case 86: /* proto_stack ::= proto_stack SLASH field */
#line 498 "mate_grammar.lemon"
{
	int* hfidp = (int *)g_malloc(sizeof(int));

	g_string_append_printf(mc->fields_filter,"||%s",yymsp[0].minor.yy202->abbrev);

	*hfidp = yymsp[0].minor.yy202->id;
	g_ptr_array_add(yymsp[-2].minor.yy231,hfidp);
	yygotominor.yy231 = yymsp[-2].minor.yy231;
  yy_destructor(yypParser,40,&yymsp[-1].minor);
}
#line 1864 "mate_grammar.c"
        break;
      case 87: /* proto_stack ::= field */
#line 508 "mate_grammar.lemon"
{
	int* hfidp = (int *)g_malloc(sizeof(int));
	*hfidp = yymsp[0].minor.yy202->id;

	g_string_append_printf(mc->fields_filter,"||%s",yymsp[0].minor.yy202->abbrev);

	yygotominor.yy231 = g_ptr_array_new();
	g_ptr_array_add(yygotominor.yy231,hfidp);
}
#line 1877 "mate_grammar.c"
        break;
      case 88: /* field ::= NAME */
#line 518 "mate_grammar.lemon"
{
	yygotominor.yy202 = proto_registrar_get_byname(yymsp[0].minor.yy0);
}
#line 1884 "mate_grammar.c"
        break;
      case 89: /* gop_decl ::= GOP_KW NAME ON_KW pdu_name MATCH_KW avpl OPEN_BRACE gop_start_statement gop_stop_statement extra_statement transform_list_statement gop_expiration_statement idle_timeout_statement lifetime_statement gop_drop_unassigned_statement show_goptree_statement show_times_statement CLOSE_BRACE SEMICOLON */
#line 536 "mate_grammar.lemon"
{
		mate_cfg_gop* cfg;

	if (g_hash_table_lookup(mc->gopcfgs,yymsp[-17].minor.yy0)) configuration_error(mc,"yygotominor.yy0 Gop Named '%s' exists already.",yymsp[-17].minor.yy0);
	if (g_hash_table_lookup(mc->gops_by_pduname,yymsp[-15].minor.yy88) ) configuration_error(mc,"Gop for Pdu '%s' exists already",yymsp[-15].minor.yy88);

	cfg = new_gopcfg(yymsp[-17].minor.yy0);
	g_hash_table_insert(mc->gops_by_pduname,yymsp[-15].minor.yy88,cfg);
	g_hash_table_insert(mc->gopcfgs,cfg->name,cfg);

	cfg->on_pdu = yymsp[-15].minor.yy88;
	cfg->key = yymsp[-13].minor.yy170;
	cfg->drop_unassigned = yymsp[-4].minor.yy216;
	cfg->show_times = yymsp[-2].minor.yy216;
	cfg->pdu_tree_mode = (gop_pdu_tree_t)yymsp[-3].minor.yy93;
	cfg->expiration = yymsp[-7].minor.yy83;
	cfg->idle_timeout = yymsp[-6].minor.yy83;
	cfg->lifetime = yymsp[-5].minor.yy83;
	cfg->start = yymsp[-11].minor.yy170;
	cfg->stop = yymsp[-10].minor.yy170;
	cfg->transforms = yymsp[-8].minor.yy231;

	merge_avpl(cfg->extra,yymsp[-9].minor.yy170,TRUE);
	delete_avpl(yymsp[-9].minor.yy170,TRUE);
  yy_destructor(yypParser,12,&yymsp[-18].minor);
  yy_destructor(yypParser,41,&yymsp[-16].minor);
  yy_destructor(yypParser,25,&yymsp[-14].minor);
  yy_destructor(yypParser,4,&yymsp[-12].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1919 "mate_grammar.c"
        break;
      case 91: /* gop_drop_unassigned_statement ::= */
#line 563 "mate_grammar.lemon"
{ yygotominor.yy216 =  mc->defaults.gop.drop_unassigned; }
#line 1924 "mate_grammar.c"
        break;
      case 92: /* gop_start_statement ::= START_KW avpl SEMICOLON */
#line 565 "mate_grammar.lemon"
{ yygotominor.yy170 = yymsp[-1].minor.yy170;   yy_destructor(yypParser,42,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1931 "mate_grammar.c"
        break;
      case 93: /* gop_start_statement ::= */
      case 95: /* gop_stop_statement ::= */ yytestcase(yyruleno==95);
#line 566 "mate_grammar.lemon"
{ yygotominor.yy170 = NULL; }
#line 1937 "mate_grammar.c"
        break;
      case 94: /* gop_stop_statement ::= STOP_KW avpl SEMICOLON */
#line 568 "mate_grammar.lemon"
{ yygotominor.yy170 = yymsp[-1].minor.yy170;   yy_destructor(yypParser,43,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1944 "mate_grammar.c"
        break;
      case 96: /* show_goptree_statement ::= SHOW_TREE_KW gop_tree_mode SEMICOLON */
#line 571 "mate_grammar.lemon"
{ yygotominor.yy93 = yymsp[-1].minor.yy93;   yy_destructor(yypParser,21,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1951 "mate_grammar.c"
        break;
      case 97: /* show_goptree_statement ::= */
#line 572 "mate_grammar.lemon"
{ yygotominor.yy93 = (gop_tree_mode_t)mc->defaults.gop.pdu_tree_mode; }
#line 1956 "mate_grammar.c"
        break;
      case 98: /* show_times_statement ::= SHOW_TIMES_KW true_false SEMICOLON */
#line 574 "mate_grammar.lemon"
{ yygotominor.yy216 = yymsp[-1].minor.yy216;   yy_destructor(yypParser,22,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1963 "mate_grammar.c"
        break;
      case 99: /* show_times_statement ::= */
#line 575 "mate_grammar.lemon"
{ yygotominor.yy216 = mc->defaults.gop.show_times; }
#line 1968 "mate_grammar.c"
        break;
      case 100: /* gop_expiration_statement ::= EXPIRATION_KW time_value SEMICOLON */
      case 118: /* gog_expiration_statement ::= EXPIRATION_KW time_value SEMICOLON */ yytestcase(yyruleno==118);
#line 577 "mate_grammar.lemon"
{ yygotominor.yy83 = yymsp[-1].minor.yy83;   yy_destructor(yypParser,18,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1976 "mate_grammar.c"
        break;
      case 101: /* gop_expiration_statement ::= */
      case 103: /* idle_timeout_statement ::= */ yytestcase(yyruleno==103);
      case 105: /* lifetime_statement ::= */ yytestcase(yyruleno==105);
#line 578 "mate_grammar.lemon"
{ yygotominor.yy83 = mc->defaults.gop.lifetime; }
#line 1983 "mate_grammar.c"
        break;
      case 102: /* idle_timeout_statement ::= IDLE_TIMEOUT_KW time_value SEMICOLON */
#line 580 "mate_grammar.lemon"
{ yygotominor.yy83 = yymsp[-1].minor.yy83;   yy_destructor(yypParser,19,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1990 "mate_grammar.c"
        break;
      case 104: /* lifetime_statement ::= LIFETIME_KW time_value SEMICOLON */
#line 583 "mate_grammar.lemon"
{ yygotominor.yy83 = yymsp[-1].minor.yy83;   yy_destructor(yypParser,20,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 1997 "mate_grammar.c"
        break;
      case 106: /* gop_tree_mode ::= NO_TREE_KW */
#line 586 "mate_grammar.lemon"
{ yygotominor.yy93 = (gop_tree_mode_t)GOP_NO_TREE;   yy_destructor(yypParser,44,&yymsp[0].minor);
}
#line 2003 "mate_grammar.c"
        break;
      case 107: /* gop_tree_mode ::= PDU_TREE_KW */
#line 587 "mate_grammar.lemon"
{ yygotominor.yy93 = (gop_tree_mode_t)GOP_PDU_TREE;   yy_destructor(yypParser,45,&yymsp[0].minor);
}
#line 2009 "mate_grammar.c"
        break;
      case 108: /* gop_tree_mode ::= FRAME_TREE_KW */
#line 588 "mate_grammar.lemon"
{ yygotominor.yy93 = (gop_tree_mode_t)GOP_FRAME_TREE;   yy_destructor(yypParser,46,&yymsp[0].minor);
}
#line 2015 "mate_grammar.c"
        break;
      case 109: /* gop_tree_mode ::= BASIC_TREE_KW */
#line 589 "mate_grammar.lemon"
{ yygotominor.yy93 = (gop_tree_mode_t)GOP_BASIC_PDU_TREE;   yy_destructor(yypParser,47,&yymsp[0].minor);
}
#line 2021 "mate_grammar.c"
        break;
      case 110: /* true_false ::= TRUE_KW */
#line 591 "mate_grammar.lemon"
{ yygotominor.yy216 = TRUE;   yy_destructor(yypParser,48,&yymsp[0].minor);
}
#line 2027 "mate_grammar.c"
        break;
      case 111: /* true_false ::= FALSE_KW */
#line 592 "mate_grammar.lemon"
{ yygotominor.yy216 = FALSE;   yy_destructor(yypParser,49,&yymsp[0].minor);
}
#line 2033 "mate_grammar.c"
        break;
      case 112: /* pdu_name ::= NAME */
#line 594 "mate_grammar.lemon"
{
	mate_cfg_pdu* c;
	if (( c =  (mate_cfg_pdu *)g_hash_table_lookup(mc->pducfgs,yymsp[0].minor.yy0) )) {
		yygotominor.yy88 = c->name;
	} else {
		configuration_error(mc,"No such Pdu: '%s'",yymsp[0].minor.yy0);
	}
}
#line 2045 "mate_grammar.c"
        break;
      case 113: /* time_value ::= FLOATING */
      case 114: /* time_value ::= INTEGER */ yytestcase(yyruleno==114);
#line 604 "mate_grammar.lemon"
{
	yygotominor.yy83 = (float) g_ascii_strtod(yymsp[0].minor.yy0,NULL);
}
#line 2053 "mate_grammar.c"
        break;
      case 115: /* gog_decl ::= GOG_KW NAME OPEN_BRACE gog_key_statements extra_statement transform_list_statement gog_expiration_statement gog_goptree_statement show_times_statement CLOSE_BRACE SEMICOLON */
#line 622 "mate_grammar.lemon"
{
		mate_cfg_gog* cfg = NULL;

		if ( g_hash_table_lookup(mc->gogcfgs,yymsp[-9].minor.yy0) ) {
			configuration_error(mc,"Gog '%s' exists already ",yymsp[-9].minor.yy0);
		}

		cfg = new_gogcfg(yymsp[-9].minor.yy0);

		cfg->expiration = yymsp[-4].minor.yy83;
		cfg->gop_tree_mode = yymsp[-3].minor.yy93;
		cfg->transforms = yymsp[-5].minor.yy231;
		cfg->keys = yymsp[-7].minor.yy243;
		cfg->show_times = yymsp[-2].minor.yy216;

		merge_avpl(cfg->extra,yymsp[-6].minor.yy170,TRUE);
		delete_avpl(yymsp[-6].minor.yy170,TRUE);
  yy_destructor(yypParser,13,&yymsp[-10].minor);
  yy_destructor(yypParser,4,&yymsp[-8].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 2079 "mate_grammar.c"
        break;
      case 116: /* gog_goptree_statement ::= GOP_TREE_KW gop_tree_type SEMICOLON */
#line 641 "mate_grammar.lemon"
{ yygotominor.yy93 = yymsp[-1].minor.yy93;   yy_destructor(yypParser,23,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 2086 "mate_grammar.c"
        break;
      case 117: /* gog_goptree_statement ::= */
#line 642 "mate_grammar.lemon"
{ yygotominor.yy93 = mc->defaults.gog.gop_tree_mode; }
#line 2091 "mate_grammar.c"
        break;
      case 119: /* gog_expiration_statement ::= */
#line 645 "mate_grammar.lemon"
{ yygotominor.yy83 = mc->defaults.gog.expiration; }
#line 2096 "mate_grammar.c"
        break;
      case 120: /* gop_tree_type ::= NULL_TREE_KW */
#line 647 "mate_grammar.lemon"
{ yygotominor.yy93 = GOP_NULL_TREE;   yy_destructor(yypParser,51,&yymsp[0].minor);
}
#line 2102 "mate_grammar.c"
        break;
      case 121: /* gop_tree_type ::= FULL_TREE_KW */
#line 648 "mate_grammar.lemon"
{ yygotominor.yy93 = GOP_FULL_TREE;   yy_destructor(yypParser,52,&yymsp[0].minor);
}
#line 2108 "mate_grammar.c"
        break;
      case 122: /* gop_tree_type ::= BASIC_TREE_KW */
#line 649 "mate_grammar.lemon"
{ yygotominor.yy93 = GOP_BASIC_TREE;   yy_destructor(yypParser,47,&yymsp[0].minor);
}
#line 2114 "mate_grammar.c"
        break;
      case 123: /* gog_key_statements ::= gog_key_statements gog_key_statement */
#line 651 "mate_grammar.lemon"
{
	loal_append(yymsp[-1].minor.yy243,yymsp[0].minor.yy170);
	yygotominor.yy243 = yymsp[-1].minor.yy243;
}
#line 2122 "mate_grammar.c"
        break;
      case 124: /* gog_key_statements ::= gog_key_statement */
#line 656 "mate_grammar.lemon"
{
	yygotominor.yy243 = new_loal("");
	loal_append(yygotominor.yy243,yymsp[0].minor.yy170);
}
#line 2130 "mate_grammar.c"
        break;
      case 125: /* gog_key_statement ::= MEMBER_KW gop_name avpl SEMICOLON */
#line 662 "mate_grammar.lemon"
{
	rename_avpl(yymsp[-1].minor.yy170,yymsp[-2].minor.yy88);
	yygotominor.yy170 = yymsp[-1].minor.yy170;
  yy_destructor(yypParser,53,&yymsp[-3].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 2140 "mate_grammar.c"
        break;
      case 126: /* gop_name ::= NAME */
#line 667 "mate_grammar.lemon"
{
	mate_cfg_gop* c;
	if (( c = (mate_cfg_gop *)g_hash_table_lookup(mc->gopcfgs,yymsp[0].minor.yy0) )) {
		yygotominor.yy88 = c->name;
	} else {
		configuration_error(mc,"No Gop called '%s' has been already declared",yymsp[0].minor.yy0);
	}
}
#line 2152 "mate_grammar.c"
        break;
      case 127: /* extra_statement ::= EXTRA_KW avpl SEMICOLON */
#line 679 "mate_grammar.lemon"
{ yygotominor.yy170 = yymsp[-1].minor.yy170;   yy_destructor(yypParser,54,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 2159 "mate_grammar.c"
        break;
      case 128: /* extra_statement ::= */
#line 680 "mate_grammar.lemon"
{ yygotominor.yy170 = new_avpl(""); }
#line 2164 "mate_grammar.c"
        break;
      case 129: /* transform_list_statement ::= TRANSFORM_KW transform_list SEMICOLON */
#line 682 "mate_grammar.lemon"
{ yygotominor.yy231 = yymsp[-1].minor.yy231;   yy_destructor(yypParser,24,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
#line 2171 "mate_grammar.c"
        break;
      case 130: /* transform_list_statement ::= */
#line 683 "mate_grammar.lemon"
{ yygotominor.yy231 = g_ptr_array_new(); }
#line 2176 "mate_grammar.c"
        break;
      case 131: /* transform_list ::= transform_list COMMA transform */
#line 685 "mate_grammar.lemon"
{
	yygotominor.yy231 = yymsp[-2].minor.yy231;
	g_ptr_array_add(yymsp[-2].minor.yy231,yymsp[0].minor.yy71);
  yy_destructor(yypParser,55,&yymsp[-1].minor);
}
#line 2185 "mate_grammar.c"
        break;
      case 132: /* transform_list ::= transform */
#line 690 "mate_grammar.lemon"
{
	yygotominor.yy231 = g_ptr_array_new();
	g_ptr_array_add(yygotominor.yy231,yymsp[0].minor.yy71);
}
#line 2193 "mate_grammar.c"
        break;
      case 133: /* transform ::= NAME */
#line 695 "mate_grammar.lemon"
{
	AVPL_Transf* t;

	if (( t = (AVPL_Transf *)g_hash_table_lookup(mc->transfs,yymsp[0].minor.yy0) )) {
		yygotominor.yy71 = t;
	} else {
		configuration_error(mc,"There's no such Transformation: %s",yymsp[0].minor.yy0);
	}
}
#line 2206 "mate_grammar.c"
        break;
      case 134: /* avpl ::= OPEN_PARENS avps CLOSE_PARENS */
#line 705 "mate_grammar.lemon"
{ yygotominor.yy170 = yymsp[-1].minor.yy170;   yy_destructor(yypParser,56,&yymsp[-2].minor);
  yy_destructor(yypParser,57,&yymsp[0].minor);
}
#line 2213 "mate_grammar.c"
        break;
      case 135: /* avpl ::= OPEN_PARENS CLOSE_PARENS */
#line 706 "mate_grammar.lemon"
{ yygotominor.yy170 = new_avpl("");   yy_destructor(yypParser,56,&yymsp[-1].minor);
  yy_destructor(yypParser,57,&yymsp[0].minor);
}
#line 2220 "mate_grammar.c"
        break;
      case 136: /* avps ::= avps COMMA avp */
#line 708 "mate_grammar.lemon"
{ yygotominor.yy170 = yymsp[-2].minor.yy170; if ( ! insert_avp(yymsp[-2].minor.yy170,yymsp[0].minor.yy110) ) delete_avp(yymsp[0].minor.yy110);   yy_destructor(yypParser,55,&yymsp[-1].minor);
}
#line 2226 "mate_grammar.c"
        break;
      case 137: /* avps ::= avp */
#line 709 "mate_grammar.lemon"
{ yygotominor.yy170 = new_avpl(""); if ( ! insert_avp(yygotominor.yy170,yymsp[0].minor.yy110) ) delete_avp(yymsp[0].minor.yy110); }
#line 2231 "mate_grammar.c"
        break;
      case 138: /* avp ::= NAME AVP_OPERATOR value */
#line 711 "mate_grammar.lemon"
{ yygotominor.yy110 = new_avp(yymsp[-2].minor.yy0,yymsp[0].minor.yy88,*yymsp[-1].minor.yy0); }
#line 2236 "mate_grammar.c"
        break;
      case 139: /* avp ::= NAME */
#line 712 "mate_grammar.lemon"
{ yygotominor.yy110 = new_avp(yymsp[0].minor.yy0,"",'?'); }
#line 2241 "mate_grammar.c"
        break;
      case 140: /* avp ::= NAME OPEN_BRACE avp_oneoff CLOSE_BRACE */
#line 713 "mate_grammar.lemon"
{ yygotominor.yy110 = new_avp(yymsp[-3].minor.yy0,yymsp[-1].minor.yy88,'|');   yy_destructor(yypParser,4,&yymsp[-2].minor);
  yy_destructor(yypParser,5,&yymsp[0].minor);
}
#line 2248 "mate_grammar.c"
        break;
      case 141: /* avp_oneoff ::= avp_oneoff PIPE value */
#line 715 "mate_grammar.lemon"
{ yygotominor.yy88 = g_strdup_printf("%s|%s",yymsp[-2].minor.yy88,yymsp[0].minor.yy88);   yy_destructor(yypParser,59,&yymsp[-1].minor);
}
#line 2254 "mate_grammar.c"
        break;
      case 142: /* avp_oneoff ::= value */
#line 716 "mate_grammar.lemon"
{ yygotominor.yy88 = g_strdup(yymsp[0].minor.yy88); }
#line 2259 "mate_grammar.c"
        break;
      case 143: /* value ::= QUOTED */
      case 144: /* value ::= NAME */ yytestcase(yyruleno==144);
      case 145: /* value ::= FLOATING */ yytestcase(yyruleno==145);
      case 146: /* value ::= INTEGER */ yytestcase(yyruleno==146);
      case 147: /* value ::= DOTED_IP */ yytestcase(yyruleno==147);
#line 718 "mate_grammar.lemon"
{ yygotominor.yy88 = g_strdup(yymsp[0].minor.yy0); }
#line 2268 "mate_grammar.c"
        break;
      case 148: /* value ::= COLONIZED */
#line 723 "mate_grammar.lemon"
{ yygotominor.yy88 = recolonize(mc,yymsp[0].minor.yy0); }
#line 2273 "mate_grammar.c"
        break;
      default:
      /* (0) mate_config ::= decls */ yytestcase(yyruleno==0);
      /* (1) decls ::= decls decl */ yytestcase(yyruleno==1);
      /* (2) decls ::= */ yytestcase(yyruleno==2);
      /* (3) decl ::= pdu_decl */ yytestcase(yyruleno==3);
      /* (4) decl ::= gop_decl */ yytestcase(yyruleno==4);
      /* (5) decl ::= gog_decl */ yytestcase(yyruleno==5);
      /* (6) decl ::= transform_decl */ yytestcase(yyruleno==6);
      /* (7) decl ::= defaults_decl */ yytestcase(yyruleno==7);
      /* (8) decl ::= debug_decl */ yytestcase(yyruleno==8);
      /* (13) dbgfile_default ::= */ yytestcase(yyruleno==13);
      /* (15) dbglevel_default ::= */ yytestcase(yyruleno==15);
      /* (17) pdu_dbglevel_default ::= */ yytestcase(yyruleno==17);
      /* (19) gop_dbglevel_default ::= */ yytestcase(yyruleno==19);
      /* (21) gog_dbglevel_default ::= */ yytestcase(yyruleno==21);
      /* (24) pdu_defaults ::= */ yytestcase(yyruleno==24);
      /* (26) pdu_last_extracted_default ::= */ yytestcase(yyruleno==26);
      /* (28) pdu_drop_unassigned_default ::= */ yytestcase(yyruleno==28);
      /* (30) pdu_discard_default ::= */ yytestcase(yyruleno==30);
      /* (32) gop_defaults ::= */ yytestcase(yyruleno==32);
      /* (34) gop_expiration_default ::= */ yytestcase(yyruleno==34);
      /* (36) gop_idle_timeout_default ::= */ yytestcase(yyruleno==36);
      /* (38) gop_lifetime_default ::= */ yytestcase(yyruleno==38);
      /* (40) gop_drop_unassigned_default ::= */ yytestcase(yyruleno==40);
      /* (42) gop_tree_mode_default ::= */ yytestcase(yyruleno==42);
      /* (44) gop_show_times_default ::= */ yytestcase(yyruleno==44);
      /* (46) gog_defaults ::= */ yytestcase(yyruleno==46);
      /* (48) gog_expiration_default ::= */ yytestcase(yyruleno==48);
      /* (50) gog_goptree_default ::= */ yytestcase(yyruleno==50);
      /* (52) gog_show_times_default ::= */ yytestcase(yyruleno==52);
        break;
/********** End reduce actions ************************************************/
  };
  assert( yyruleno>=0 && yyruleno<(int)(sizeof(yyRuleInfo)/sizeof(yyRuleInfo[0])) );
  yygoto = yyRuleInfo[yyruleno].lhs;
  yysize = yyRuleInfo[yyruleno].nrhs;
  yypParser->yyidx -= yysize;
  yyact = yy_find_reduce_action(yymsp[-yysize].stateno,(YYCODETYPE)yygoto);
  if( yyact <= YY_MAX_SHIFTREDUCE ){
    if( yyact>YY_MAX_SHIFT ) yyact += YY_MIN_REDUCE - YY_MIN_SHIFTREDUCE;
    /* If the reduce action popped at least
    ** one element off the stack, then we can push the new element back
    ** onto the stack here, and skip the stack overflow test in yy_shift().
    ** That gives a significant speed improvement. */
    if( yysize ){
      yypParser->yyidx++;
      yymsp -= yysize-1;
      yymsp->stateno = (YYACTIONTYPE)yyact;
      yymsp->major = (YYCODETYPE)yygoto;
      yymsp->minor = yygotominor;
      yyTraceShift(yypParser, yyact);
    }else{
      yy_shift(yypParser,yyact,yygoto,&yygotominor);
    }
  }else{
    assert( yyact == YY_ACCEPT_ACTION );
    yy_accept(yypParser);
  }
}

/*
** The following code executes when the parse fails
*/
#ifndef YYNOERRORRECOVERY
static void yy_parse_failed(
  yyParser *yypParser           /* The parser */
){
  MateParserARG_FETCH;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sFail!\n",yyTracePrompt);
  }
#endif
  while( yypParser->yyidx>=0 ) yy_pop_parser_stack(yypParser);
  /* Here code is inserted which will be executed whenever the
  ** parser fails */
/************ Begin %parse_failure code ***************************************/
#line 190 "./mate_grammar.lemon"

	configuration_error(mc,"Parse Error");
#line 2355 "mate_grammar.c"
/************ End %parse_failure code *****************************************/
  MateParserARG_STORE; /* Suppress warning about unused %extra_argument variable */
}
#endif /* YYNOERRORRECOVERY */

/*
** The following code executes when a syntax error first occurs.
*/
static void yy_syntax_error(
  yyParser *yypParser,           /* The parser */
  int yymajor _U_,               /* The major type of the error token */
  YYMINORTYPE yyminor            /* The minor type of the error token */
){
  MateParserARG_FETCH;
#define TOKEN (yyminor.yy0)
/************ Begin %syntax_error code ****************************************/
#line 186 "./mate_grammar.lemon"

	configuration_error(mc,"Syntax Error before %s",yyminor);
#line 2375 "mate_grammar.c"
/************ End %syntax_error code ******************************************/
  MateParserARG_STORE; /* Suppress warning about unused %extra_argument variable */
}

/*
** The following is executed when the parser accepts
*/
static void yy_accept(
  yyParser *yypParser           /* The parser */
){
  MateParserARG_FETCH;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sAccept!\n",yyTracePrompt);
  }
#endif
  while( yypParser->yyidx>=0 ) yy_pop_parser_stack(yypParser);
  /* Here code is inserted which will be executed whenever the
  ** parser accepts */
/*********** Begin %parse_accept code *****************************************/
/*********** End %parse_accept code *******************************************/
  MateParserARG_STORE; /* Suppress warning about unused %extra_argument variable */
}

/* The main parser program.
** The first argument is a pointer to a structure obtained from
** "MateParserAlloc" which describes the current state of the parser.
** The second argument is the major token number.  The third is
** the minor token.  The fourth optional argument is whatever the
** user wants (and specified in the grammar) and is available for
** use by the action routines.
**
** Inputs:
** <ul>
** <li> A pointer to the parser (an opaque structure.)
** <li> The major token number.
** <li> The minor token number.
** <li> An option argument of a grammar-specified type.
** </ul>
**
** Outputs:
** None.
*/
void MateParser(
  void *yyp,                   /* The parser */
  int yymajor,                 /* The major token code number */
  MateParserTOKENTYPE yyminor       /* The value for the token */
  MateParserARG_PDECL               /* Optional %extra_argument parameter */
){
  YYMINORTYPE yyminorunion;
  int yyact;            /* The parser action. */
#if !defined(YYERRORSYMBOL) && !defined(YYNOERRORRECOVERY)
  int yyendofinput;     /* True if we are at the end of input */
#endif
#ifdef YYERRORSYMBOL
  int yyerrorhit = 0;   /* True if yymajor has invoked an error */
#endif
  yyParser *yypParser;  /* The parser */

  /* (re)initialize the parser, if necessary */
  yypParser = (yyParser*)yyp;
  if( yypParser->yyidx<0 ){
#if YYSTACKDEPTH<=0
    if( yypParser->yystksz <=0 ){
      /*memset(&yyminorunion, 0, sizeof(yyminorunion));*/
      yyminorunion = yyzerominor;
      yyStackOverflow(yypParser, &yyminorunion);
      return;
    }
#endif
    yypParser->yyidx = 0;
    yypParser->yyerrcnt = -1;
    yypParser->yystack[0].stateno = 0;
    yypParser->yystack[0].major = 0;
#ifndef NDEBUG
    if( yyTraceFILE ){
      fprintf(yyTraceFILE,"%sInitialize. Empty stack. State 0\n",
              yyTracePrompt);
    }
#endif
  }
  yyminorunion.yy0 = yyminor;
#if !defined(YYERRORSYMBOL) && !defined(YYNOERRORRECOVERY)
  yyendofinput = (yymajor==0);
#endif
  MateParserARG_STORE;

#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sInput '%s'\n",yyTracePrompt,yyTokenName[yymajor]);
  }
#endif

  do{
    yyact = yy_find_shift_action(yypParser,(YYCODETYPE)yymajor);
    if( yyact <= YY_MAX_SHIFTREDUCE ){
      if( yyact > YY_MAX_SHIFT ) yyact += YY_MIN_REDUCE - YY_MIN_SHIFTREDUCE;
      yy_shift(yypParser,yyact,yymajor,&yyminorunion);
      yypParser->yyerrcnt--;
      yymajor = YYNOCODE;
    }else if( yyact <= YY_MAX_REDUCE ){
      yy_reduce(yypParser,yyact-YY_MIN_REDUCE);
    }else{
      assert( yyact == YY_ERROR_ACTION );
#ifdef YYERRORSYMBOL
      int yymx;
#endif
#ifndef NDEBUG
      if( yyTraceFILE ){
        fprintf(yyTraceFILE,"%sSyntax Error!\n",yyTracePrompt);
      }
#endif
#ifdef YYERRORSYMBOL
      /* A syntax error has occurred.
      ** The response to an error depends upon whether or not the
      ** grammar defines an error token "ERROR".
      **
      ** This is what we do if the grammar does define ERROR:
      **
      **  * Call the %syntax_error function.
      **
      **  * Begin popping the stack until we enter a state where
      **    it is legal to shift the error symbol, then shift
      **    the error symbol.
      **
      **  * Set the error count to three.
      **
      **  * Begin accepting and shifting new tokens.  No new error
      **    processing will occur until three tokens have been
      **    shifted successfully.
      **
      */
      if( yypParser->yyerrcnt<0 ){
        yy_syntax_error(yypParser,yymajor,yyminorunion);
      }
      yymx = yypParser->yystack[yypParser->yyidx].major;
      if( yymx==YYERRORSYMBOL || yyerrorhit ){
#ifndef NDEBUG
        if( yyTraceFILE ){
          fprintf(yyTraceFILE,"%sDiscard input token %s\n",
             yyTracePrompt,yyTokenName[yymajor]);
        }
#endif
        yy_destructor(yypParser, (YYCODETYPE)yymajor,&yyminorunion);
        yymajor = YYNOCODE;
      }else{
         while(
          yypParser->yyidx >= 0 &&
          yymx != YYERRORSYMBOL &&
          (yyact = yy_find_reduce_action(
                        yypParser->yystack[yypParser->yyidx].stateno,
                        YYERRORSYMBOL)) >= YY_MIN_REDUCE
        ){
          yy_pop_parser_stack(yypParser);
        }
        if( yypParser->yyidx < 0 || yymajor==0 ){
          yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
          yy_parse_failed(yypParser);
          yymajor = YYNOCODE;
        }else if( yymx!=YYERRORSYMBOL ){
          YYMINORTYPE u2;
          u2.YYERRSYMDT = 0;
          yy_shift(yypParser,yyact,YYERRORSYMBOL,&u2);
        }
      }
      yypParser->yyerrcnt = 3;
      yyerrorhit = 1;
#elif defined(YYNOERRORRECOVERY)
      /* If the YYNOERRORRECOVERY macro is defined, then do not attempt to
      ** do any kind of error recovery.  Instead, simply invoke the syntax
      ** error routine and continue going as if nothing had happened.
      **
      ** Applications can set this macro (for example inside %include) if
      ** they intend to abandon the parse upon the first syntax error seen.
      */
      yy_syntax_error(yypParser,yymajor,yyminorunion);
      yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
      yymajor = YYNOCODE;

#else  /* YYERRORSYMBOL is not defined */
      /* This is what we do if the grammar does not define ERROR:
      **
      **  * Report an error message, and throw away the input token.
      **
      **  * If the input token is $, then fail the parse.
      **
      ** As before, subsequent error messages are suppressed until
      ** three input tokens have been successfully shifted.
      */
      if( yypParser->yyerrcnt<=0 ){
        yy_syntax_error(yypParser,yymajor,yyminorunion);
      }
      yypParser->yyerrcnt = 3;
      yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
      if( yyendofinput ){
        yy_parse_failed(yypParser);
      }
      yymajor = YYNOCODE;
#endif
    }
  }while( yymajor!=YYNOCODE && yypParser->yyidx>=0 );
#ifndef NDEBUG
  if( yyTraceFILE ){
    int i;
    fprintf(yyTraceFILE,"%sReturn. Stack=",yyTracePrompt);
    for(i=1; i<=yypParser->yyidx; i++)
      fprintf(yyTraceFILE,"%c%s", i==1 ? '[' : ' ',
              yyTokenName[yypParser->yystack[i].major]);
    fprintf(yyTraceFILE,"]\n");
  }
#endif
  return;
}
